"""Object Relational Model for merge-queue database."""

import enum
import json

import sqlalchemy
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import DateTime
from sqlalchemy import ForeignKey

GERRIT_TIME_SHORT_FMT = '%Y-%m-%d %H:%M:%S'

# A sqlalchemy concept, a kind of 'registry' of the SQL object mapping
Base = declarative_base()  # pylint: disable=invalid-name

# constants used to indicate build results


class StatusKey(enum.Enum):
  TIMEOUT = -3
  CANCELED = -2
  STEP_FAILED = -1
  SUCCESS = 0
  IN_PROGRESS = 1


class Cancellation(Base):   # pylint: disable=no-init
  """
  Whether or not a merge has been cancedled
  """

  __tablename__ = 'cancellations'
  __table_args__ = {'sqlite_autoincrement': True}

  # row/record id of merge to cancel
  rid = Column(Integer, primary_key=True)

  # who canceled
  who = Column(String)

  # when the cancellation happened
  when = Column(DateTime)


class Builder(Base):  # pylint: disable=no-init
  """
  Information about servers available to build changes
  """
  __tablename__ = 'builders'
  __table_args__ = {'sqlite_autoincrement': True}

  # row/record id
  rid = Column(Integer, primary_key=True)

  # human readable / friendly name
  name = Column(String)

  # shared secret authentication key
  auth_key = Column(String)

  # Additional non-indexed information about this builder. In particular
  # contains the configuration for each project/branch build
  builder_meta = Column(String)

  def __repr__(self):
    return ('<Builder(id="{}", name="{}">'
            .format(self.rid, self.name))

  def as_dict(self):
    result = {key: getattr(self, key) for key
              in ['rid', 'name']}
    return result


class MergeStatusV0p1p0(Base):  # pylint: disable=no-init
  """
  Stores information about a merge attempt. This structure is serialized
  into a row of the SQLite database storing the merge history.
  """
  __tablename__ = 'merge_history_v0p1p0'
  __table_args__ = {'sqlite_autoincrement': True}

  id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name
  gerrit_id = Column(String)
  change_id = Column(String)
  feature_branch = Column(String)
  target_branch = Column(String)
  owner = Column(String)
  request_time = Column(DateTime)
  start_time = Column(DateTime)
  end_time = Column(DateTime)
  result = Column(Integer)

  def __repr__(self):
    return ('<MergeResult(id="{}")>').format(self.id)


class MergeStatusV0p2p0(Base):  # pylint: disable=no-init
  """
  Data about an attempt to complete a merge.
  """
  __tablename__ = 'merge_history_v0p2p0'
  __table_args__ = {'sqlite_autoincrement': True}

  # row/record id
  rid = Column(Integer, primary_key=True)

  # the name of the project
  project = Column(String, index=True)

  # the name of the target branch
  branch = Column(String, index=True)

  # rid of AccountInfo table
  owner_id = Column(Integer, ForeignKey('account_info.rid'))
  owner = relationship('AccountInfo')

  # gerrit change id
  change_id = Column(String)

  # time on the gerrit server of the first MQ+1 after the last MQ-1
  request_time = Column(DateTime)

  # time that the daemon actually started the merge
  start_time = Column(DateTime)

  # time that the daemon completed the merge
  end_time = Column(DateTime)

  # status of the merge. See values in the StatusKey enum.
  status = Column(Integer)

  # progress in fractions of 1/10,000. i.e. 10,000 means 100% done.
  progress = Column(Integer)

  # Additional non-indexed information as a json object. Includes things
  # like feature-branch, target-branch, etc.
  msg_meta = Column(String)

  def __repr__(self):
    return ('<MergeAttempt(id="{}", gerrit_id="{}/{}/{}">'
            .format(self.rid, self.project, self.branch, self.change_id))

  def as_dict(self):
    result = {key: getattr(self, key) for key
              in ['rid', 'project', 'branch', 'change_id', 'status',
                  'progress']}
    for key in ['request_time', 'start_time', 'end_time']:
      result[key] = getattr(self, key).strftime(GERRIT_TIME_SHORT_FMT)

    if self.msg_meta is None:
      result['metadata'] = {}
    else:
      result['metadata'] = json.loads(self.msg_meta)

    if hasattr(self, 'owner') and isinstance(self.owner, AccountInfo):
      result['owner'] = self.owner.as_dict()
    else:
      result['owner'] = {'rid': self.owner_id,
                         'name': '<unknown>',
                         'email': '<unknown>',
                         'username': '<unknown>'}

    return result


class MergeStatus(Base):  # pylint: disable=no-init
  """
  Data about an attempt to complete a merge.
  """
  __tablename__ = 'merge_history'
  __table_args__ = {'sqlite_autoincrement': True}

  # row/record id
  rid = Column(Integer, primary_key=True)

  # the name of the project
  project = Column(String, index=True)

  # the name of the target branch
  branch = Column(String, index=True)

  # time that the daemon actually started the merge
  start_time = Column(DateTime)

  # time that the daemon completed the merge
  end_time = Column(DateTime)

  # status of the merge. See values in the StatusKey enum.
  status = Column(Integer)

  def __repr__(self):
    return ('<MergeStatus(id="{}", gerrit_id="{}/{}">'
            .format(self.rid, self.project, self.branch))

  def as_dict(self):
    result = {key: getattr(self, key) for key
              in ['rid', 'project', 'branch', 'status']}

    for key in ['start_time', 'end_time']:
      result[key] = getattr(self, key).strftime(GERRIT_TIME_SHORT_FMT)

    return result


class MergeChange(Base):  # pylint: disable=no-init
  """
  Information about one change that was part of one merge attempt
  """

  __tablename__ = 'merge_changes'
  __table_args__ = {'sqlite_autoincrement': True}

  # row/record id
  rid = Column(Integer, primary_key=True)

  # row/record id of the merge that this change was a part of
  merge_id = Column(Integer, ForeignKey('merge_history.rid'))
  merge = relationship("MergeStatus")

  # rid of AccountInfo table
  owner_id = Column(Integer, ForeignKey('account_info.rid'))
  owner = relationship('AccountInfo')

  # gerrit change id
  change_id = Column(String)

  # time on the gerrit server of the first MQ+1 after the last MQ-1
  request_time = Column(DateTime)

  # extracted 'Feature-Branch' from message headers
  feature_branch = Column(String)

  # Additional non-indexed information as a json object. Includes things
  # like feature-branch, target-branch, etc.
  msg_meta = Column(String)

  def __repr__(self):
    return ('<MergeChange(id="{}", gerrit_id="{}">'
            .format(self.rid, self.change_id))

  def as_dict(self):
    result = {key: getattr(self, key) for key
              in ['rid', 'change_id', 'merge_id', 'owner_id']}

    for key in ['request_time']:
      result[key] = getattr(self, key).strftime(GERRIT_TIME_SHORT_FMT)

    if self.msg_meta is None:
      result['metadata'] = {}
    else:
      result['metadata'] = json.loads(self.msg_meta)

    if hasattr(self, 'owner') and isinstance(self.owner, AccountInfo):
      result['owner'] = self.owner.as_dict()
    else:
      result['owner'] = {'rid': self.owner_id}

    if hasattr(self, 'merge') and isinstance(self.merge, MergeStatus):
      result['merge'] = self.merge.as_dict()
    else:
      result['merge'] = {'rid': self.merge_id}

    return result


class AccountInfo(Base):  # pylint: disable=no-init
  """
  Local cache of gerrit AccoutnInfo objects  to reduce the number of gerrit
  REST hits that we have to make.
  """

  __tablename__ = 'account_info'

  # record id, matches _account_id from gerrit AccountInfo
  rid = Column(Integer, primary_key=True)

  # full name of the user
  name = Column(String)

  # email address of the user
  email = Column(String)

  # email address of the user
  username = Column(String)

  def __repr__(self):
    return ('<AccountInfo(id="{}", username="{}">'
            .format(self.rid, self.username))

  def as_dict(self):
    result = {key: getattr(self, key) for key
              in ['rid', 'name', 'email', 'username']}
    result['_account_id'] = self.rid
    return result


class ChangeInfo(Base):  # pylint: disable=no-init
  """
  Local cache of gerrit ChangeInfo objects to reduce the number of gerrit
  REST hits that we have to make.
  """
  __tablename__ = 'change_queue'
  __table_args__ = {'sqlite_autoincrement': True}

  # row/record id
  rid = Column(Integer, primary_key=True)

  # the name of the project
  project = Column(String, index=True)

  # the name of the target branch
  branch = Column(String, index=True)

  # gerrit change id
  change_id = Column(String)

  # header line of the commit message
  subject = Column(String)

  # current revision of the change
  current_revision = Column(String)

  # id of the owner AccountInfo (in account_info table)
  owner_id = Column(Integer, ForeignKey('account_info.rid'))
  owner = relationship('AccountInfo')

  # json-encoded dictionary of any metadata stored in colon-tags in the
  # commit message.
  message_meta = Column(String)

  # time on the gerrit server of the first MQ+1 after the last MQ-1
  queue_time = Column(DateTime)

  # resolved queue score after evaluating label order
  queue_score = Column(Integer)

  # unique identifier for when this change info was cached
  poll_id = Column(Integer, index=True)

  # merge priority. 0 is highest priority, 100 is default priority. Lower
  # value has higher priority.
  priority = Column(Integer, index=True)

  def __repr__(self):
    return ('<ChangeInfo(change_id="{}/{}/{}")>'
            .format(self.project, self.branch, self.change_id))

  def as_dict(self):
    result = {key: getattr(self, key) for key
              in ['rid', 'project', 'branch', 'change_id', 'subject',
                  'current_revision', 'queue_score', 'poll_id', 'priority']}
    result['owner'] = self.owner.as_dict()
    result['queue_time'] = self.queue_time.strftime(GERRIT_TIME_SHORT_FMT)
    if self.message_meta is None:
      result['message_meta'] = {}
    else:
      result['message_meta'] = json.loads(self.message_meta)
    return result

  @classmethod
  def from_dict(cls, poll_id, project, branch,  # pylint: disable=W0613,R0913
                change_id, subject, current_revision, owner, queue_time,
                queue_score, message_meta=None, **kwargs):
    if message_meta is None:
      meta_str = ''
    else:
      meta_str = json.dumps(message_meta)

    return cls(project=project,
               branch=branch,
               change_id=change_id,
               subject=subject,
               current_revision=current_revision,
               owner_id=owner.get('_account_id', -1),
               message_meta=meta_str,
               queue_time=queue_time,
               queue_score=queue_score,
               poll_id=poll_id,
               priority=message_meta.get('Priority', 100))


class QueueSpec(Base):  # pylint: disable=no-init
  """
  Specification/Configuration for a serialization queue
  """

  __tablename__ = 'queue_spec'
  __table_args__ = {'sqlite_autoincrement': True}

  # row/record id
  rid = Column(Integer, primary_key=True)

  # The project this queue applies to
  project = Column(String, index=True)

  # Name of this queue (same as branch if this is not a regex)
  name = Column(String)

  # Regular expression for branch names to match
  branch = Column(String)

  # JSON encoded dictionary to use as the environment
  build_env = Column(String)

  # Whether or not the host environment should be merged into the build
  # environment
  merge_build_env = Column(Boolean)

  # Json endcoded list of commands to execute for the build
  build_steps = Column(String)

  # Submit with rest api. If false, the last step in `build_steps` should
  # do the submit
  submit_with_rest = Column(Boolean)

  def __repr__(self):
    return ('<QueueSpec({}, {})>'
            .format(self.project, self.name))

  def as_dict(self):
    result = {key: getattr(self, key) for key
              in ['rid', 'project', 'name', 'branch', 'merge_build_env',
                  'submit_with_rest']}
    result['build_env'] = json.loads(self.build_env)
    result['build_steps'] = json.loads(self.build_steps)
    return result


def init_sql(database_url):
  """
  Initialize sqlalchemy and the sqlite database. Returns a session factory.
  """
  engine = sqlalchemy.create_engine(database_url, echo=False)
  Base.metadata.create_all(engine)
  return sqlalchemy.orm.sessionmaker(bind=engine)



"""Provides common functionality for unattended merge tools."""

import datetime
import json
import logging
import urllib

import pygerrit2.rest
import requests


class ConfigDict(dict):
  """
  Wraps a nested native python dictionary with semantics for accessing
  nested elements through decimal delimited paths.
  """

  def __getitem__(self, key):
    if (isinstance(key, str) or isinstance(key, unicode)) and '.' in key:
      path = key.split('.')
      sub = self
      for part in path:
        sub = sub[part]
      return sub
    else:
      return super(ConfigDict, self).__getitem__(key)

  def __setitem__(self, key, value):
    if (isinstance(key, str) or isinstance(key, unicode)) and '.' in key:
      path = key.split('.')
      sub = self
      for part in path[:-1]:
        if part not in sub:
          sub[part] = {}
        sub = sub[part]
      return sub.__setitem__(path[-1], value)
    else:
      return super(ConfigDict, self).__setitem__(key, value)

  def __contains__(self, key):
    if (isinstance(key, str) or isinstance(key, unicode)) and '.' in key:
      path = key.split('.')
      sub = self
      for part in path:
        if part in sub:
          sub = sub[part]
        else:
          return False
      return True
    else:
      return super(ConfigDict, self).__contains__(key)

  def get(self, key, default=None):
    if (isinstance(key, str) or isinstance(key, unicode)) and '.' in key:
      path = key.split('.')
      sub = self
      for part in path:
        if part in sub:
          sub = sub[part]
        else:
          return default
      return sub
    else:
      return super(ConfigDict, self).get(key, default)


def is_valid_changeinfo(json_dict):
  """
  Simple sanity check on a ChangeInfo object. Ensures that it at least
  contains enough information to do our job.
  """
  for key in ['id', 'branch', 'change_id', 'owner', 'updated',
              'current_revision', 'status', 'subject']:
    if key not in json_dict:
      return False
  return True


GERRIT_TIME_SHORT_FMT = '%Y-%m-%d %H:%M:%S'
GERRIT_TIME_FMT = '%Y-%m-%d %H:%M:%S.%f'


def sort_merge_queue_labels(label_entries):
  """
  Given a list of label entries, find all the entries where a merge queue
  score was added, parse the date for that label entry, then sort them
  by date.

  Returns a list of (time, score) tuples.
  """

  timestamped_list = []
  for label_entry in label_entries:
    # if a user sets the value of the Merge-Queue label to 0 the date is
    # removed
    if 'date' not in label_entry:
      continue
    if 'value' not in label_entry:
      continue

    # strip trailing zeros so that strptime doesn't complain
    label_date_str = label_entry['date'].rstrip('0')

    # but if we've stripped all the way to the dot, we've gone too far,
    # and strptime will complain
    if label_date_str.endswith('.'):
      label_date_str += '0'

    label_time = datetime.datetime.strptime(label_date_str, GERRIT_TIME_FMT)
    timestamped_list.append((label_time, label_entry['value']))

  return sorted(timestamped_list)


def get_resolved_merge_queue_score(sorted_labels):
  """
  Given a list of (time, label) pairs, find earliest '+1' after the latest '-1',
  if it exists, and return (time, label). Otherwise return (now, '-1').
  """

  resolved_time = datetime.datetime.utcnow()
  resolved_score = -1

  for mq_time, mq_score in sorted_labels:
    if mq_score == -1:
      resolved_time = mq_time
      resolved_score = mq_score
    elif mq_score == 1 and resolved_score != 1:
      resolved_time = mq_time
      resolved_score = mq_score

  return resolved_time, resolved_score


def gerrit_query(filters):
  """
  Format a query string given gerrit query filters. The query string is
  composed of url-encoded space ('+') separated key:value pairs. The value
  will be quoted if it contains whitespace like key:"value x".
  """
  pairs = []
  for key, value in sorted(filters):
    value = urllib.quote_plus(value)
    if '+' in value:
      pairs.append('{}:"{}"'.format(key, value))
    else:
      pairs.append('{}:{}'.format(key, value))
  return '+'.join(pairs)


class AccountInfo(object):  # pylint: disable=no-init
  """
  Information about a gerrit user
  """

  def __init__(self, _account_id, username, name, email,
               rid=None):  # pylint:disable=unused-argument
    self.account_id = _account_id
    self.username = username
    self.name = name
    self.email = email

  def as_dict(self):
    result = {key: getattr(self, key) for key
              in ['name', 'email', 'username']}
    result['_account_id'] = self.account_id
    return result


class ChangeInfo(object):
  """
  Information about a gerrit change
  """

  def __init__(self, project, branch,  # pylint: disable=unused-argument
               change_id, subject, current_revision, owner, queue_time,
               queue_score, message_meta=None, **kwargs):
    self.project = project
    self.branch = branch
    self.change_id = change_id
    self.subject = subject
    self.current_revision = current_revision
    self.owner = AccountInfo(**owner)
    if message_meta is None:
      self.message_meta = {}
    else:
      self.message_meta = dict(message_meta)
    if isinstance(queue_time, datetime.datetime):
      self.queue_time = queue_time
    elif isinstance(queue_time, str) or isinstance(queue_time, unicode):
      self.queue_time = datetime.datetime.strptime(queue_time,
                                                   GERRIT_TIME_SHORT_FMT)
    else:
      raise ValueError('Unrecognized queue_time type of {}'
                       .format(type(queue_time)))
    self.queue_score = queue_score

  def as_dict(self):
    result = {key: getattr(self, key) for key
              in ['project', 'branch', 'subject', 'current_revision', 'owner',
                  'message_meta', 'change_id', 'queue_time']}
    result['owner'] = self.owner.as_dict()
    result['queue_time'] = self.queue_time.strftime(GERRIT_TIME_SHORT_FMT)
    return result

  def pretty_string(self):
    try:
      import yaml
      return yaml.safe_dump(self.as_dict(), indent=2, width=80)
    except ImportError:
      return json.dumps(self.as_dict(), indent=2, sort_keys=True)

  @staticmethod
  def key(changeinfo):

    return (changeinfo.message_meta.get('Priority', 100),
            changeinfo.queue_time, changeinfo.project,
            changeinfo.change_id)


class GerritRest(pygerrit2.rest.GerritRestAPI):
  """
  Provides a high-level interface to the Gerrit REST API for the few specific
  queries we need. This is shared between the unattende merge daemon and the
  web front-end so that the web front-end see's the same queue as the daemon.
  """

  def __init__(self, url, username, password,
               disable_ssl_certificate_validation=False):
    auth = requests.auth.HTTPDigestAuth(username, password)
    verify = (not disable_ssl_certificate_validation)
    super(GerritRest, self).__init__(url=url, auth=auth, verify=verify)

    if disable_ssl_certificate_validation:
      from requests.packages import urllib3
      urllib3.disable_warnings()

  def get_merge_requests(self, offset=0, limit=25, filters=None):
    """
    Call out the gerrit REST API and return a list of all changes that are
    marked as being requested for merge, in the correct queue-order.

    filters : a list of (key, value) pairs to filter for. For instance
              [('project', 'project_foo')] would limit the query to only
              'project_foo'.

    The following filters are automatically appended:
      * status:new
      * label:code-review=+2
      * label:merge-queue=+1

    returns: a list of (queue_time, changinfo) pairs
    """

    # this is the query that we send to gerrit. We want a list of all changes
    # which are
    #   1) 'NEW' status
    #   2) have been code-reviewed with +2
    #   3) have been labeled with merge-queue +1
    # NOTE(josh): status: 'open' include 'new' and 'submitted', where
    # 'submitted' means that gerrit has not yet merged it. We want to exclude
    # 'submitted'.
    if filters is None:
      filters = []

    filters += [('status', 'new'),
                ('label', 'code-review=+2'),
                ('label', 'merge-queue=+1')]

    # these are extra outputs that we want as part of the query. For each
    # change we want a list of labels that have been assigned, as well as the
    # current revision for the change. The current revision is required for
    # the follow-up query to get the feature branch name
    query_string = ('q=' + gerrit_query(filters) + '&'
                    + urllib.urlencode([('o', 'CURRENT_REVISION'),
                                        ('o', 'LABELS'),
                                        ('o', 'DETAILED_LABELS'),
                                        ('o', 'DETAILED_ACCOUNTS'),
                                        ('start', offset),
                                        ('n', limit)]))

    try:
      parsed_changes = self.get('changes/?' + query_string)
    except (requests.RequestException, ValueError):
      logging.exception('Failed to query queue from gerrit.')
      logging.error('query was:\n' + query_string)
      return []

    changeinfo_list = []
    for json_dict in parsed_changes:
      if not is_valid_changeinfo(json_dict):
        logging.error('Invalid JSON ChangeInfo:')
        logging.error(json.dumps(json_dict, sort_keys=True, indent=2,
                                 separators=(',', ': ')))
        continue

      mq_labels = (json_dict
                   .get('labels', {})
                   .get('Merge-Queue', {})
                   .get('all', []))
      sorted_labels = sort_merge_queue_labels(mq_labels)
      queue_time, queue_score = get_resolved_merge_queue_score(sorted_labels)

      if queue_score == 1:
        commit_message_meta = self.get_message_meta(
            json_dict['change_id'], json_dict['current_revision'])
        json_dict['message_meta'] = commit_message_meta
        json_dict['queue_time'] = queue_time
        json_dict['queue_score'] = queue_score
        changeinfo_list.append(ChangeInfo(**json_dict))
      else:
        logging.info('Skipping change %s with resolved '
                     'Merge-Queue label of %d',
                     json_dict['change_id'], queue_score)

    if len(changeinfo_list) == 0:
      return []

    # NOTE(josh): Gerrit returns the results most-recently-touched-first, so
    # we sort by the timestamp of the earliest +1 after the latest -1 on that
    # change
    return sorted(changeinfo_list, key=ChangeInfo.key)

  def get_changeinfo(self, change_id):
    """
    Return the ChangeInfo object for a particular change id
    """

    query = urllib.urlencode([('o', 'CURRENT_REVISION'),
                              ('o', 'LABELS'),
                              ('o', 'DETAILED_LABELS'),
                              ('o', 'DETAILED_ACCOUNTS')])
    query_uri = 'changes/{}?{}'.format(change_id, query)
    json_dict = self.get(query_uri)

    if not is_valid_changeinfo(json_dict):
      logging.error('Invalid JSON ChangeInfo:')
      logging.error(json.dumps(json_dict, sort_keys=True, indent=2,
                               separators=(',', ': ')))
      return None
    return json_dict

  def get_change(self, change_id):
    json_dict = self.get_changeinfo(change_id)

    mq_labels = (json_dict
                 .get('labels', {})
                 .get('Merge-Queue', {})
                 .get('all', []))
    sorted_labels = sort_merge_queue_labels(mq_labels)
    queue_time, queue_score = get_resolved_merge_queue_score(sorted_labels)

    json_dict['queue_time'] = queue_time
    json_dict['queue_score'] = queue_score
    return ChangeInfo(**json_dict)

  def get_changes_canceled_on_gerrit(self, change_queue):
    """
    Retrieve updated changeinfo objects for each change in change_queue and
    return a list of ``change_id`` objects for all changes that have been
    canceled.
    """

    changes_str = ' OR '.join('change:{}'.format(changeinfo.change_id)
                              for changeinfo in change_queue)
    search_query = ("project:{project} AND branch:{branch} AND ({change_list})"
                    .format(project=change_queue[0].project,
                            branch=change_queue[0].branch,
                            change_list=changes_str))
    query_string = urllib.urlencode([('q', search_query),
                                     ('o', 'DETAILED_LABELS')])
    canceled_ids = []
    for changeinfo in self.get('changes/?' + query_string):
      mq_labels = (changeinfo
                   .get('labels', {})
                   .get('Merge-Queue', {})
                   .get('all', []))
      sorted_labels = sort_merge_queue_labels(mq_labels)
      _, queue_score = get_resolved_merge_queue_score(sorted_labels)
      if queue_score != 1:
        canceled_ids.append(changeinfo['change_id'])
    return canceled_ids

  def get_message_meta(self, change_id, revision):
    """
    Call out to gerrit REST API to get the commit message for the most recent
    revision of a a change, and then scan the commit message for the metadata
    tags. Return a dictionary of metadata found.

    The following metadata keys are handled specially:
      * Closes
      * Resolves

    They may be written more than once in the commit message and the contents
    will be merged. The contents are expected to be a comma separated list of
    strings. The output dictionary will contain the separated list of strings.
    """

    parsed_details = self.get('changes/{}/revisions/{}/commit'
                              .format(change_id, revision))

    if 'message' not in parsed_details:
      raise RuntimeError('Message is not a field in returned json')

    result = dict(Closes=[], Resolves=[])
    for line in parsed_details['message'].splitlines():
      parts = line.split(':', 1)
      if len(parts) == 2:
        key, value = parts
        if key in ['Closes', 'Resolves']:
          issues = [item.strip() for item in value.strip().split(',')]
          result[key].extend(issues)
        elif key in ['Priority']:
          try:
            result[key] = int(value)
          except ValueError:
            pass
        else:
          result[key] = value.strip()

    return result

  def submit_change(self, change_id, author_id=None):
    request_url = 'changes/{}/submit'.format(change_id)

    try:
      if author_id is not None:
        return self.post(request_url, json={'on_behalf_of': author_id})
      else:
        return self.post(request_url)

    except requests.RequestException:
      logging.exception('Failed to set review score for change %s', change_id)
      return None

  def set_review(self, change_id, current_revision, review_dict):
    """
    Call out to the gerrit REST API to set a review comment on the specified
    change. The format for the review_dict is that of a ReviewInput json
    object.

    See: https://gerrit-review.googlesource.com/Documentation/
         rest-api-changes.html#review-input.
    """
    request_url = 'changes/{}/revisions/{}/review'.format(change_id,
                                                          current_revision)
    try:
      self.post(request_url, json=review_dict)
    except requests.RequestException:
      logging.exception('Failed to set review score for change %s', change_id)

  def get_username_from_email(self, email_address):
    """
    Call out to the gerrit REST API to get the user info associated with an
    email address, and return the gerrit username for that email address.
    """
    request_url = 'accounts/{}'.format(email_address)
    try:
      parsed_content = self.get(request_url)
      return parsed_content.get('username', None)
    except requests.RequestException:
      logging.exception('Failed to get username for email %s', email_address)
      return None



"""
Merge-Queue daemon, repeatedly polls gerrit for any merge requests, and merges
them in serial order.
"""

import datetime
import json
import logging.handlers
import os
import re
import signal
import subprocess
import time

import git
import httplib2
import requests

from gerrit_mq import orm
from gerrit_mq import functions

# TODO(josh): split this module
# pylint: disable=too-many-lines

IN_SUBMISSION_TPL = """
Gerrit Merge-Queue has started to merge this change as part of merge #{1}.
{0}/detail.html?merge_id={1}
"""

RESULT_TPL = """
Merge #{1} {2}.
{0}/detail.html?merge_id={1}
"""

FAILURE_TPL = """

********************************
Merge failed on step {stepno}. The following command exited with nonzero status:
{command}

The return code was {retcode}
********************************

"""

STEP_TPL = """

-------------------------
Executing step: {stepno}
{command}
-------------------------

"""

GERRIT_CANCEL = """

********************************
Merge canceled because the following changes were evicted on gerrit by score
removal:
%s
********************************

"""

WEBFRONT_CANCEL = """

********************************
Merge was canceled through the webfront
  by: %s
  on: %s
********************************

"""


class QueueSpec(object):
  """
  Specification of a single queue of serialized merges
  """

  def __init__(self, project, branch, build_env, build_steps, name=None,
               merge_build_env=False, submit_with_rest=True, coalesce_count=0,
               submit_cmd=None):
    self.project = project
    self.branch = re.compile(branch)
    self.build_env = dict(build_env)
    self.build_steps = [list(step) for step in build_steps]
    self.coalesce_count = coalesce_count

    # Changes that were part of a failed coaleced verification. As long as
    # one of these changes is part of the current queue head, changes will
    # be made in serial order.
    self.dirty_changes = set()

    if name is None:
      assert re.escape(branch) == branch
      self.name = branch
    else:
      self.name = name

    self.merge_build_env = merge_build_env
    self.submit_with_rest = submit_with_rest
    if submit_cmd is None:
      self.submit_cmd = []
    else:
      self.submit_cmd = submit_cmd

    for key in self.build_env:
      value = self.build_env[key]

      # Environment variables like PATH or PYTHONPATH may be specified as
      # lists or tuples, in which case we join the elements with a
      # pathseparator to make the path string.
      if key.endswith('PATH') and (isinstance(value, list)
                                   or isinstance(value, tuple)):
        self.build_env[key] = ':'.join(value)

  def get_workspace(self, base_path):
    """
    Return the repo directory for this queue
    """

    return os.path.join(base_path, self.project, self.name)

  def get_environment(self, config):
    """
    Return the environment used to execute commands for this queue
    """

    if config.get('daemon.merge_build_env', False):
      subenv = os.environ.copy()
      subenv.update(self.build_env)
    else:
      subenv = dict(self.build_env)

    if 'daemon.ccache.path' in config:
      subenv['CCACHE_DIR'] = config['daemon.ccache.path']
    return subenv


def mark_gerrit_change_as_in_submission(gerrit, changeinfo,
                                        webfront_url, merge_rid):
  """
  Posts a review to a gerrit change informing owner and reviewers that it is
  being merged.
  """
  message = IN_SUBMISSION_TPL.format(webfront_url, merge_rid)
  review_dict = {'message': message,
                 'labels': {'Merge-Queue': 0},
                 'notify': 'NONE'}  # don't email on merge started
  gerrit.set_review(changeinfo.change_id, changeinfo.current_revision,
                    review_dict)


def get_result_message(webfront_url, merge_rid, merge_result):
  """
  Get the message to post to gerrit on build completion
  """

  if merge_result == orm.StatusKey.SUCCESS.value:
    result_string = 'successful'
  elif merge_result == orm.StatusKey.CANCELED.value:
    result_string = 'cancelled'
  else:
    result_string = 'failed'

  return RESULT_TPL.format(webfront_url, merge_rid, result_string)


def mark_gerrit_change_with_result(gerrit, changeinfo, webfront_url, merge_rid,
                                   merge_result):
  """
  Posts a review to a gerrit change informing owner and reviewers the merge
  succeeded or failed.
  """

  message = get_result_message(webfront_url, merge_rid, merge_result)

  if merge_result == 0:
    label = 1
  else:
    label = -1

  review_dict = {'message': message,
                 'labels': {'Merge-Queue': label}}
  # If the merge succeeds the user will already get an email from gerrit
  # so there's no need to email again
  if merge_result == 0:
    review_dict['notify'] = 'NONE'
  gerrit.set_review(changeinfo.change_id, changeinfo.current_revision,
                    review_dict)


def fetch_branches_from_origin(repo):
    # fetch branches from origin
  logging.info('Fetching branches from origin')
  repo.git.fetch('origin', prune=True)


def merge_a_into_b(repo, branch_a, branch_b):
  logging.info('Checking out %s', branch_b)
  repo.git.checkout(branch_b)
  logging.info('Merging %s into %s', branch_a, branch_b)

  # NOTE(josh): aN is 'author name' and aE is 'author email'.
  author_str = repo.git.show('HEAD', no_patch=True,
                             format="%aN <%aE>").strip()
  author = '"{}"'.format(author_str)

  # NOTE(justin): adding this instrumentation to understand why this merge
  # step sometimes seems to no-op when it should take action.
  target_head = repo.git.show(branch_a, no_patch=True, format="%h").strip()
  logging.info('%s head commit: %s', branch_a, target_head)

  # NOTE(josh): I verified that even with --no-commit specified, the merge
  # exits with error code 1, so this command should raise an exception if the
  # merge is not clean
  repo.git.merge(branch_a, no_commit=True)

  # This is a trick to use the default merge commit message. We set the
  # editor to a command which ignores it's arguments and exits with error
  # code 0. Git will generate a merge message, write it to a temporary file,
  # then call the editor with that file as a command line argument. The
  # command 'true' immediately exits with error code '0' and the commit
  # message is unchanged. Note that 'true' is an actual binary executable
  # in /bin/true on ubuntu 14.04.
  old_env = repo.git.update_environment(GIT_EDITOR='true')

  try:
    # Attempt to commit the merge.  If the merge was a no-op, this commit
    # will fail, and we can just ignore that failure. We pass --no-verify
    # because we don't want to run pre-commit hooks for this merge.
    logging.info('Committing merge')
    repo.git.commit(author=author, no_verify=True)
  except git.exc.GitCommandError:
    # NOTE(justin): This commit seems to be failing sometimes for reasons
    # besides being a no-op, i.e. it's actually a failed commit.  If it really
    # was a no-op we should have  an empty git_status; otherwise re-raise the
    # error we received.
    git_status = repo.git.status(porcelain=True)
    if git_status:
      logging.info('Merge commit failed, but git status is not clean. '
                   'Git status:\n%s', git_status)
      raise
    else:
      logging.info('Merge was a no-op; skipping commit.')

  # Restore the git environment as it was before our little hack.
  repo.git.update_environment(**old_env)


def merge_features_together(repo, merge_branch, change_queue):
  """
  Create a new branch, merge all the changes into it the amature way.
  """
  target_branch = change_queue[0].branch
  logging.info('Checking out target branch %s', target_branch)
  repo.git.checkout(target_branch)

  logging.info('Creating merge branch %s', merge_branch)
  new_branch = repo.create_head(merge_branch)
  new_branch.checkout()

  for changeinfo in change_queue:
    feature_branch = changeinfo.message_meta['Feature-Branch']
    # merge target into feature
    merge_a_into_b(repo, merge_branch, feature_branch)
    # merge updated feature back into target
    merge_a_into_b(repo, feature_branch, merge_branch)


def kill_step(step_proc):
  logging.info('Waiting for build step to die, pid=%d', step_proc.pid)
  start_time = time.time()

  if step_proc.returncode is None:
    logging.info('Signalling with SIGTERM')
  while step_proc.poll() is None:
    step_proc.send_signal(signal.SIGTERM)
    time.sleep(2)
    if time.time() - start_time > 10:
      break

  if step_proc.returncode is None:
    logging.info('Signalling with SIGKILL')

  while step_proc.poll() is None:
    step_proc.send_signal(signal.SIGKILL)
    time.sleep(2)
    if time.time() - start_time > 10:
      break

  logging.info('Build step appears to be zombified. Hopefully it does not'
               'affect future merges')


def mark_old_changes_as_failed(sql):
  """
  If the daemon was killed during a merge, then mark that merge as failed.
  """

  query = (sql.query(orm.MergeStatus)
           .filter(orm.MergeStatus.status == orm.StatusKey.IN_PROGRESS.value))
  for merge_status in query:
    logging.info('Marking stale merge status %d as failed', merge_status.rid)
    merge_status.status = orm.StatusKey.CANCELED.value
  sql.commit()


def submit_changes_with_rest(gerrit, change_queue):
  """
  Submit the list of changes through the gerrit REST api
  """
  for changeinfo in change_queue:
    logging.info('Submitting %s through REST API',
                 changeinfo.change_id)
    # NOTE(josh): on-behalf-of appears to be restricted with our current
    # configuration. Otherwise use changeinfo.owner.account_id or the
    # account_id of whoever supplied the resolved mergequeue score
    response = gerrit.submit_change(changeinfo.change_id)
    if response.get('status') != 'SUBMITTED':
      logging.warn('Gerrit refused to submit the change over REST')
      break


def submit_changes_with_cmd(repo, change_queue, submit_cmd, popen_kwargs):
  """
  Submit the list of changes using a command.
  """
  target_branch = change_queue[0].branch
  for changeinfo in change_queue:
    logging.info('Checking out target branch: %s', target_branch)
    repo.git.checkout(target_branch)

    logging.info('Pulling target branch state from gerrit')
    repo.git.pull()

    feature_branch = changeinfo.message_meta['Feature-Branch']
    merge_a_into_b(repo, target_branch, feature_branch)

    logging.info('Submitting %s through command line',
                 changeinfo.change_id)
    try:
      step_proc = subprocess.Popen(submit_cmd, **popen_kwargs)
    except OSError:
      logging.exception("Failed to execute %s", ' '.join(submit_cmd))
      break

    while step_proc.poll() is None:
      # TODO(josh): timeout here if submit takes too long
      time.sleep(1)


def create_sql_records(sql, queue_spec, change_queue):
  """
  Create a record for the merge including records for each change verified as
  part of this merge verification. Return the main merge record.
  """

  # NOTE(josh): row id will be assigned by the database and retrieved by
  # SQLAlchemy when we commit() below.
  merge = orm.MergeStatus(
      project=queue_spec.project,
      branch=change_queue[0].branch,
      start_time=datetime.datetime.utcnow(),
      end_time=datetime.datetime.utcnow(),
      status=orm.StatusKey.IN_PROGRESS.value)
  sql.add(merge)
  sql.commit()

  for changeinfo in change_queue:
    feature_branch = changeinfo.message_meta.get('Feature-Branch', None)
    if feature_branch is None:
      raise RuntimeError('No Feature-Branch in message for {}'
                         .format(changeinfo.change_id))
    change = orm.MergeChange(
        merge_id=merge.rid,
        change_id=changeinfo.change_id,
        owner_id=changeinfo.owner.account_id,
        feature_branch=feature_branch,
        request_time=changeinfo.queue_time,
        msg_meta=json.dumps(changeinfo.message_meta))
    sql.add(change)
  sql.commit()

  return merge


def run_steps(queue_spec, gerrit, change_queue, sql_session, merge_id,
              popen_kwargs):
  """
  Performs each build, test step.
  """

  logging.info('Performing build/test steps')

  for step_idx, step_cmd in enumerate(queue_spec.build_steps):
    # Reset every step so we check at least once per step
    last_gerrit_poll = 0
    last_db_poll = 0
    last_timing_print = 0
    step_start_time = time.time()

    # Write the command that we are running for this step into the log so we
    # can associate stdout and stderr with the command that was run
    for stream in ['stdout', 'stderr']:
      log = popen_kwargs[stream]
      log.write(STEP_TPL.format(stepno=step_idx, command=' '.join(step_cmd)))
      log.flush()

    try:
      step_proc = subprocess.Popen(step_cmd, **popen_kwargs)
    except OSError:
      logging.exception("Failed to execute %s", ' '.join(step_cmd))
      raise

    command_str = ' '.join(step_cmd)
    logging.info('{} {}'.format(step_idx, command_str))

    gerrit_poll_count = 0
    gerrit_poll_failures = 0

    if queue_spec.submit_with_rest:
      should_poll_gerrit = True
    else:
      if step_idx < len(queue_spec.build_steps) - 1:
        should_poll_gerrit = True
      else:
        # NOTE(josh): if this is the last of the build steps and the merge queue
        # does not submit through the REST api, then this step must do the
        # actual merge somehow (i.e. through the gerrit REST api or through the
        # gerrit command line interface). In this case it may alter the state
        # of the review which may remove the MQ+1 score which we SHOULD NOT
        # interpret as a merge request cancellation.
        should_poll_gerrit = False

    while step_proc.poll() is None:
      # Print a message every two minutes for sanity
      if time.time() - last_timing_print > 5 * 60:
        last_timing_print = time.time()
        step_duration = last_timing_print - step_start_time
        logging.debug('Step %d has been running for %6.2f seconds',
                      step_idx, step_duration)

      # NOTE(josh): check for cancellation on gerrit every 30 seconds
      if should_poll_gerrit and (time.time() - last_gerrit_poll > 30):
        last_gerrit_poll = time.time()
        canceled_ids = []
        try:
          gerrit_poll_count += 1
          canceled_ids = gerrit.get_changes_canceled_on_gerrit(change_queue)
        except (requests.RequestException, ValueError):
          gerrit_poll_failures += 1
          if gerrit_poll_failures == 1:
            logging.exception("Failed to poll gerrit for changes.\n"
                              " NOTE(josh): This is known to happen from time "
                              " to time, so don't be too concerned.")
          else:
            logging.warn('Failed to poll gerrit for changes %d/%d',
                         gerrit_poll_failures, gerrit_poll_count)
          continue

        if canceled_ids:
          logging.info(GERRIT_CANCEL, '\n  '.join(canceled_ids))
          kill_step(step_proc)
          return orm.StatusKey.CANCELED.value

      # NOTE(josh): check for cancellation on mq database
      if time.time() - last_db_poll > 10:
        last_db_poll = time.time()

        query = (sql_session.query(orm.Cancellation)
                 .filter(orm.Cancellation.rid == merge_id))
        for record in query:
          logging.info(WEBFRONT_CANCEL, record.who, record.when)
          kill_step(step_proc)
          return orm.StatusKey.CANCELED.value

      # TODO(josh): check for cancellation, update status/heartbeat,
      # check for timeout, print animation, estimate progess based on lines
      # of output, etc
      time.sleep(1)

    logging.info('{} {} [{}] '.format(step_idx, command_str,
                                      step_proc.returncode))
    if step_proc.returncode != 0:
      message = FAILURE_TPL.format(stepno=step_idx,
                                   retcode=step_proc.returncode,
                                   command=command_str)
      raise RuntimeError(message)

  return orm.StatusKey.SUCCESS.value


def cleanup_repo(repo):
  """
  Cleanup branches and working tree after merge
  """
  # clean up repo in case anything failed
  repo.git.reset('--hard')
  # -f is 'force', -d is 'remove whole directories'
  repo.git.clean('-fd')
  repo.git.checkout('master')
  repo.git.clean('-fd')

  # delete all branches except master
  for branch in repo.git.branch().split():
    if branch != '*' and branch.strip() != 'master':
      logging.info('Deleting left-over branch %s', branch)
      try:
        repo.git.branch('-D', branch)
      except git.exc.GitCommandError:
        logging.exception(
            'Failed to delete leftover feature branch %s', branch)


def get_or_clone_repo(config, repo_path, project):
  try:
    return git.Repo(repo_path)
  except (git.NoSuchPathError, git.InvalidGitRepositoryError):
    logging.exception('workspace does not appear to be a git repository: %s',
                      repo_path)

  try:
    os.makedirs(os.path.dirname(repo_path))
  except OSError:
    pass

  # clone the repository if needed
  repository_url = 'ssh://{}@{}:{}/{}.git'.format(
      config['gerrit.ssh.username'],
      config['gerrit.ssh.host'],
      config['gerrit.ssh.port'],
      project)
  logging.info('attemping to clone %s', repository_url)

  # TODO(josh): replace with subprocess call, gitpython appears to supress
  # the command output which would be pretty handy for sanity sake
  git.Repo.clone_from(repository_url, repo_path)

  # if this is a fresh clone we need to run git-fat init, so do it every time
  # anyway
  subprocess.call(['git-fat', 'init'], cwd=repo_path)

  # create a git repository object for this working repository
  return git.Repo(repo_path)


def handle_pid_file(pidfile_path):
  """
  Verify that we are the only daemon running by writing a PID file. If the
  file already exists read it to see what other daemon PID file is running.
  If that pid is no longer active assume it has died and we are allowed to
  start.
  """

  try:
    os.makedirs(os.path.dirname(pidfile_path))
  except OSError:
    pass

  other_pid = None
  try:
    with open(pidfile_path, 'r') as infile:
      other_pid = int(infile.read().strip())
  except (OSError, IOError, ValueError):
    pass

  # NOTE(josh): if we restart due to file change then this file will exist
  # but it will contain our pid.
  if other_pid is not None and other_pid != os.getpid():
    if os.path.exists('/proc/{}/stat'.format(other_pid)):
      logging.error('Another daemon is already running with pid %d.',
                    other_pid)
      return 1
    else:
      logging.warn('Daemon pid file %s exists containing pid %d which is not'
                   ' alive, will overwrite', pidfile_path, other_pid)

  with open(pidfile_path, 'w') as pidfile:
    pidfile.write('{}\n'.format(os.getpid()))


def get_requests_matching(request_queue, project, branch):
  """
  Filter request_queue returning a list of only those requests matching the
  given branch and project
  """

  return [cinfo for cinfo in request_queue
          if cinfo.project == project and cinfo.branch == branch]


def get_requests_from_single_queue(request_queue, queue_specs):
  """
  Find the first merge request in `request_queue` that matches a specification
  in `queue_specs`. Then, given that `queue_spec`, build and return a list of
  all outstanding changes to that (`project`, `branch`).
  """
  for cinfo in request_queue:
    if cinfo.project in queue_specs:
      for spec in queue_specs[cinfo.project]:
        if spec.branch.match(cinfo.branch):
          return spec, get_requests_matching(request_queue, cinfo.project,
                                             cinfo.branch)
  return None, []


class LogInfo(object):

  def __init__(self):
    self.app_logpath = None
    self.log_handler = None
    self.stdout_logpath = None
    self.stdout = None
    self.stderr_logpath = None
    self.stderr = None


def setup_logs(log_path, merge_id):
  """
  Open three log files for app, stdout, and stderr.
  """

  # Create a temporary logging handler which copies log events to the named
  # log file for this merge
  app_logpath = '{}/{:06d}.log'.format(log_path, merge_id)
  log_handler = logging.FileHandler(app_logpath, 'w')
  log_handler.setLevel(logging.DEBUG)
  logging.getLogger('').addHandler(log_handler)

  # Create log files for stdout and stderr of build steps
  stdout_logpath = '{}/{:06d}.stdout'.format(log_path, merge_id)
  stdout_log = open(stdout_logpath, 'w')

  stderr_logpath = '{}/{:06d}.stderr'.format(log_path, merge_id)
  stderr_log = open(stderr_logpath, 'w')

  out = LogInfo()
  out.app_logpath = app_logpath
  out.log_handler = log_handler
  out.stdout = stdout_log
  out.stdout_logpath = stdout_logpath
  out.stderr = stderr_log
  out.stderr_logpath = stderr_logpath

  return out


class MergeDaemon(object):

  def __init__(self, config, gerrit, sql_session):
    self.config = config
    self.gerrit = gerrit
    self.sql_session = sql_session

    try:
      os.makedirs(config['daemon.workspace_path'])
    except OSError:
      pass

    queue_index = {}
    for spec_dict in self.config['queues']:
      spec = QueueSpec(**spec_dict)
      queue_index[(spec.project, spec.name)] = spec

    self.queues = {}
    for project, name in self.config['daemon.queues']:
      spec = queue_index.get((project, name), None)
      if spec is None:
        logging.warn('daemon queue not listed in queue index: %s/%s',
                     project, name)
        continue
      if project not in self.queues:
        self.queues[project] = []
      self.queues[project].append(spec)

    # increase ccache size
    sub_env = os.environ.copy()
    sub_env['CCACHE_DIR'] = self.config['daemon.ccache.path']

    try:
      os.makedirs(config['daemon.ccache.path'])
    except OSError:
      pass

    subprocess.check_call(['ccache', '-M', config['daemon.ccache.size']],
                          env=sub_env, cwd=config['daemon.workspace_path'])

  def coalesce_merge(self, queue_spec, change_queue):
    """
    Merge all changes from `change_queue` together, verify the build and, if
    it passes, then submit all of the changes through gerrit.
    """

    # Take this opportunity to to update the AccountInfo table with any new
    # owner info contained in this change
    for changeinfo in change_queue:
      functions.add_or_update_account_info(self.sql_session,
                                           changeinfo.owner)
      self.sql_session.commit()

    # Create a log entry for this merge attempt. Note that the id will be
    # assigned by sqlalchemy after we 'commit' to the database.
    merge = create_sql_records(self.sql_session, queue_spec, change_queue)

    silent = self.config.get('daemon.silent', False)
    logctx = setup_logs(self.config['log_path'], merge.rid)

    logging.info('Starting verification of the following changes: \n  %s',
                 '\n  '.join([changeinfo.change_id for changeinfo
                              in change_queue]))
    repo = None
    try:
      repo_path = queue_spec.get_workspace(self.config['daemon.workspace_path'])
      repo = get_or_clone_repo(self.config, repo_path=repo_path,
                               project=queue_spec.project)

      if not silent:
        message = IN_SUBMISSION_TPL.format(self.config['webfront.url'],
                                           merge.rid)
        review_dict = {'message': message,
                       'labels': {'Merge-Queue': 0},
                       'notify': 'NONE'}  # don't email on merge started
        for changeinfo in change_queue:
          self.gerrit.set_review(changeinfo.change_id,
                                 changeinfo.current_revision, review_dict)

      fetch_branches_from_origin(repo)
      merge_branch = 'mergequeue_{:06d}'.format(merge.rid)
      merge_features_together(repo, merge_branch, change_queue)

      if not silent:
        # Push the updated feature branch back to origin so its state there is
        # up to date.
        repo.git.push('origin', '{0}:{0}'.format(merge_branch), force=True)

      popen_kwargs = {
          'env': queue_spec.get_environment(self.config),
          'cwd': queue_spec.get_workspace(self.config['daemon.workspace_path']),
          'stdout': logctx.stdout,
          'stderr': logctx.stderr,
      }

      merge.status = run_steps(queue_spec, self.gerrit, change_queue,
                               self.sql_session, merge.rid, popen_kwargs)

      if not silent:
        repo.git.push('origin', ':{}'.format(merge_branch))

    except (OSError, RuntimeError, KeyError, git.exc.GitCommandError):
      merge.status = orm.StatusKey.STEP_FAILED.value
      logging.exception('Exception caught during merge')

    if repo is not None:
      cleanup_repo(repo)

    if merge.status == orm.StatusKey.SUCCESS.value:
      if queue_spec.submit_with_rest:
        submit_changes_with_rest(self.gerrit, change_queue)
      else:
        cleanup_repo(repo)
        submit_changes_with_cmd(repo, change_queue, queue_spec.submit_cmd,
                                popen_kwargs)

    # Add a comment to gerrit indicating success or failure, and setting a
    # review score for the Merge-Queue label.
    if not silent:
      message = get_result_message(self.config['webfront.url'], merge.rid,
                                   merge.status)

      # NOTE(josh): if this is the second pass, then we want the label to
      # be -1: on failure
      review_score = 0
      if (len(change_queue) == 1
          and merge.status != orm.StatusKey.SUCCESS.value):
        review_score = -1

      review_dict = {'message': message,
                     'labels': {'Merge-Queue': review_score}}

      # If the merge succeeds the user will already get an email from gerrit
      # so there's no need to email again.
      if merge.status == orm.StatusKey.SUCCESS.value:
        review_dict['notify'] = 'NONE'

      for changeinfo in change_queue:
        self.gerrit.set_review(changeinfo.change_id,
                               changeinfo.current_revision, review_dict)

    # mark the time when the merge was completed / failed
    merge.end_time = datetime.datetime.utcnow()

    # commit change to history database
    self.sql_session.commit()

    # remove the the handler that is logging messages to the file for this merge
    logging.getLogger('').removeHandler(logctx.log_handler)
    logctx.log_handler.close()
    logctx.stdout.close()
    logctx.stderr.close()

    # compress the logs
    for logpath in [logctx.app_logpath, logctx.stdout_logpath,
                    logctx.stderr_logpath]:
      subprocess.call(['gzip', '--force', logpath])

      # NOTE(josh): the nginx gzip_static module wants the original files around
      # or it wont serve the compressed ones. It's not great to rely on nginx
      # behavior but for now we touch the file to make nginx happy and then
      # hope that it wont be served.
      with open(logpath, 'w') as _:
        pass

    if merge.status == orm.StatusKey.SUCCESS.value:
      for changeinfo in change_queue:
        queue_spec.dirty_changes.discard(changeinfo.change_id)
      return 0
    else:
      for changeinfo in change_queue:
        queue_spec.dirty_changes.add(changeinfo.change_id)
      return -1

  def run(self, watch_manifest):
    pidfile_path = self.config.get('daemon.pidfile_path', './pid')
    handle_pid_file(pidfile_path)
    poll_period = self.config.get('daemon.poll_period', 60)
    offline_sentinel_path = self.config.get('daemon.offline_sentinel_path',
                                            './pause')

    mark_old_changes_as_failed(self.sql_session)
    last_poll_time = 0

    while True:
      functions.restart_if_modified(watch_manifest, pidfile_path)

      try:
        if os.path.exists(offline_sentinel_path):
          logging.info('Offline sentinal exists, bypassing merges')
          while os.path.exists(offline_sentinel_path):
            functions.restart_if_modified(watch_manifest, pidfile_path)
            time.sleep(1)
          logging.info('Offline sentinel removed, continuing')
          continue

        # If the loop was faster than poll period, then wait for
        # the remainder of the period to prevent spamming gerrit
        loop_duration = time.time() - last_poll_time
        backoff_duration = poll_period - loop_duration
        if backoff_duration > 0:
          logging.info('Loop was very fast, waiting for '
                       '%6.2f seconds', backoff_duration)
          time.sleep(backoff_duration)

        last_poll_time = time.time()
        poll_id = functions.get_next_poll_id(self.sql_session)
        functions.poll_gerrit(self.gerrit, self.sql_session, poll_id)
        _, global_queue = functions.get_queue(self.sql_session)

        queue_spec, request_queue = \
            get_requests_from_single_queue(global_queue, self.queues)

        if queue_spec is None or not request_queue:
          # If there are no changes to any of the queues that this daemon is
          # monitoring then we have nothing to do here.
          continue

        if queue_spec.coalesce_count > 0 and len(request_queue) > 1:
          # NOTE(josh): Only coalesce changes that have never failed
          # verification before.
          coalesce_queue = []
          for changeinfo in request_queue:
            if changeinfo.change_id in queue_spec.dirty_changes:
              logging.info('ceasing merge colation since %s is dirty',
                           changeinfo.change_id)
              break
            else:
              coalesce_queue.append(changeinfo)
            if len(coalesce_queue) >= queue_spec.coalesce_count:
              break

          if len(coalesce_queue) > 1:
            result = self.coalesce_merge(queue_spec, coalesce_queue)
            if result == 0:
              # The coalition of changes was verified together, they have all
              # been merged so we can poll gerrit and move on to more changes.
              continue
            else:
              for changeinfo in coalesce_queue:
                queue_spec.dirty_changes.add(changeinfo.change_id)
          else:
            logging.info('falling back to single-merge since coalition '
                         'contains only one clean change')
        else:
          logging.info('skipping merge coalition, coalesce_count: %d, '
                       'len(request_queue): %d', queue_spec.coalesce_count,
                       len(request_queue))

        # NOTE(josh): only do one merge per request to gerrit so that
        # any changes to the queue (i.e. gerrit state through review
        # updates or priority changes) are reflected in the merge order,
        # as well as allowing us to pick-up on the pause sentinel
        self.coalesce_merge(queue_spec, request_queue[:1])
        queue_spec.dirty_changes.discard(request_queue[0].change_id)

      except (httplib2.HttpLib2Error, requests.RequestException):
        logging.exception('Error retrieving merge requests from gerrit')
        continue

      except KeyboardInterrupt:
        break

    logging.info('Exiting main loop')

    return 0



"""Web frontend for inspecting the merge queue."""

from __future__ import print_function
import cStringIO

import flask
import sqlalchemy

from gerrit_mq import functions
from gerrit_mq import orm

HTML_ESCAPE_TABLE = {
    "&": "&amp;",
    '"': "&quot;",
    "'": "&apos;",
    ">": "&gt;",
    "<": "&lt;",
}


def html_escape_file(infile, max_bytes):
  """
  Escape special characters so that the content is html-embeddable
  """
  buf = cStringIO.StringIO()
  bytes_read = 0
  for char in iter(lambda: infile.read(1).decode('utf-8', 'replace'), ''):
    # TODO(josh): fix this in a better way
    if ord(char) < 128:
      buf.write(HTML_ESCAPE_TABLE.get(char, char))
    else:
      # TODO(josh): is this OK for utf-8?
      buf.write(char)
    bytes_read += 1
    if bytes_read >= max_bytes:
      break

  # TODO(josh): Can we force this to be unicode? Create test case.
  return bytes_read, buf.getvalue()


def html_escape_tail(filename, max_bytes):
  """
  Escape special characters so that the content is html-embeddable
  """
  with open(filename, 'r') as infile:
    infile.seek(0, 2)  # move the file stream to the end of the file
    size_in_bytes = infile.tell()
    if max_bytes > size_in_bytes:
      infile.seek(0, 0)  # move back to the beginning
    else:
      infile.seek(-max_bytes, 2)
    _, content = html_escape_file(infile, max_bytes)
    return content


def extract_common_args(req_args):
  """
  Pull common query parameter arguments out of a query string
  """

  # TODO(josh): figure out FlaskSqlAlchemy and how to do this right.
  project_filter = req_args.get('project', None)
  branch_filter = req_args.get('branch', None)

  try:
    offset = max(int(req_args['offset']), 0)
  except (KeyError, ValueError):
    offset = 0

  try:
    limit = min(int(req_args['limit']), 500)
  except (KeyError, ValueError):
    limit = 25

  return project_filter, branch_filter, offset, limit


class Master(flask.Flask):

  def __init__(self, mq_config, gerrit, sql_factory):
    super(Master, self).__init__('gerrit_mq')
    self.debug = mq_config['webfront.flask_debug']
    self.gerrit = gerrit
    self.sql_factory = sql_factory
    self.mq_config = mq_config
    self.secret_key = mq_config['webfront.secret_key']

    self.add_url_rule('/gmq/get_history', self.get_job)
    self.add_url_rule('/gmq/get_job', self.get_job)
    self.add_url_rule('/gmq/get_queue', self.get_queue)
    self.add_url_rule('/gmq/get_status', self.get_job)

  def get_queue(self):
    """
    Return json-encoded list of ChangeInfo objects for the current queue.

    Query params:
    `project` : SQL `LIKE` expression for projects to match
    `branch` : SQL `LIKE` expression for branches to match
    `offset` : start offset for pagination
    `limit` : maximum number of records to return
    """

    project_filter, branch_filter, offset, limit \
        = extract_common_args(flask.request.args)

    sql = self.sql_factory()
    result = functions.get_queue(sql, project_filter, branch_filter, offset,
                                 limit)
    sql.close()
    return flask.jsonify(result)

  def get_history(self):
    """
    Return json-encoded list of MergeStatus objects.

    Query params:
      `project` : SQL `LIKE` expression for projects to match
      `branch` : SQL `LIKE` expression for branches to match
      `offset` : start offset for pagination
      `limit` : maximum number of records to return
    """
    project_filter, branch_filter, offset, limit \
        = extract_common_args(flask.request.args)

    sql = self.sql_factory()
    result = functions.get_history(sql, project_filter, branch_filter, offset,
                                   limit)
    sql.close()
    return flask.jsonify(result)

  def get_status(self):
    """
    Return json-encoded MergeStatus for a single merge

    Query params:
      `rid` : row id of the status to retrieve
    """
    sql = self.sql_factory()
    try:
      query_rid = int(flask.request.args['rid'])
    except (KeyError, ValueError):
      response = flask.jsonify({'status': 'ERROR',
                                'reason': 'invalid rid'})
      response.status_code = 400
      sql.close()
      return response

    query = sql.query(orm.MergeStatus).filter(orm.MergeStatus.rid == query_rid)
    for ms_sql in query:
      response = flask.jsonify(ms_sql.as_dict())
      sql.close()
      return response

    sql.close()
    response = flask.jsonify({'status': 'ERROR',
                              'reason': "rid doesn't exist in db"})
    response.status_code = 404
    return response

  def get_job(self):
    """
    Return the next available job for a given executor

    Query params:
      `builder_id` : the builder id
      `auth_key` : the builder secret
    """
    builder = None
    builder_id = flask.request.args.get('builder_id', None)

    for builder_config in self.mq_config['builders']:
      if builder_config['id'] == builder_id:
        builder = builder_config
        break

    if builder is None:
      response = flask.jsonify({'status': 'ERROR',
                                'reason':
                                'Builder not found {}'.format(builder_id)})
      response.status_code = 404
      return response

    if builder['auth_key'] != flask.request.args.get('auth_key', None):
      response = flask.jsonify({'status': 'ERROR',
                                'reason': 'Invalid credentials'})
      response.status_code = 403
      return response

    sql = self.sql_factory()
    query = (sql
             .query(sqlalchemy.func.min(orm.ChangeInfo.query_time))
             .group_by(orm.ChangeInfo.project, orm.ChangeInfo.branch)
             .order_by(orm.ChangeInfo.query_time.desc()))

    builder_queues = []
    for ci_sql in query:
      if (ci_sql.project, ci_sql.branch) in builder_queues:
        response = flask.jsonify(ci_sql.as_dict())
        sql.close()
        return response

    sql.close()
    return flask.jsonify({})


def main(mq_config, gerrit, session_factory):  # pylint: disable=unused-argument
  app = Master(mq_config, gerrit, session_factory)
  app.run(**mq_config['webfront.listen'])



from __future__ import print_function
import ctypes
import json
import logging
import os
import shutil
import subprocess
import sys
import time
import zipfile

import jinja2
import requests
from gerrit_mq import common
from gerrit_mq import orm


def add_or_update_account_info(sql, ai_obj):
  """
  Update the AccountInfo object from gerrit json, or create it if it's new.
  """
  query = (sql
           .query(orm.AccountInfo)
           .filter(orm.AccountInfo.rid == ai_obj.account_id)
           .limit(1))
  if query.count() > 0:
    ai_sql = query.first()
    for field in ['name', 'email', 'username']:
      setattr(ai_sql, field, getattr(ai_obj, field))
  else:
    kwargs = ai_obj.as_dict()
    kwargs['rid'] = kwargs.pop('_account_id')
    for key in ['name', 'email', 'username']:
      if key not in kwargs:
        kwargs[key] = '<none>'
    ai_sql = orm.AccountInfo(**kwargs)
    sql.add(ai_sql)


def get_next_poll_id(sql):
  """
  Return the next unused poll id
  """

  from sqlalchemy.sql.expression import func
  query = sql.query(func.max(orm.ChangeInfo.poll_id))
  last_poll_id = query.scalar()
  if last_poll_id is None:
    return 1
  else:
    return last_poll_id + 1


def poll_gerrit(gerrit, sql, poll_id):
  """
  Hit gerrit REST and read off the current queue of merge requests. Update the
  local cache database entries for any changes that have been updated since
  our last poll. Write the resulting ordered queue to the queue file.
  """

  request_queue = gerrit.get_merge_requests()
  for changeinfo in request_queue:
    # Take this opportunity to to update the AccountInfo table
    # with the owner info
    add_or_update_account_info(sql, changeinfo.owner)

    priority = changeinfo.message_meta.get('Priority', 100)
    ci_sql = orm.ChangeInfo(project=changeinfo.project,
                            branch=changeinfo.branch,
                            change_id=changeinfo.change_id,
                            subject=changeinfo.subject,
                            current_revision=changeinfo.current_revision,
                            owner_id=changeinfo.owner.account_id,
                            message_meta=json.dumps(changeinfo.message_meta),
                            queue_time=changeinfo.queue_time,
                            poll_id=poll_id,
                            priority=priority)
    sql.add(ci_sql)
  sql.commit()

  # Delete anything in the cache which did not show up during this poll
  (sql.query(orm.ChangeInfo)
   .filter(orm.ChangeInfo.poll_id != poll_id)
   .delete())
  sql.commit()


def get_queue(sql, project_filter=None, branch_filter=None,
              offset=None, limit=None):
  """
  Return list of ChangeInfo objects matching the given project and branch
  filters (as SQL LIKE expressions)

  Returns a tuple of (`count`, `result_list`) where `count` is the size of
  the query without `offset` or `limit`.

  TODO(josh): filter out any which are IN_PROGRESS?
  """
  query = sql.query(orm.ChangeInfo)

  if project_filter is not None:
    query = query.filter(orm.ChangeInfo.project.like(project_filter))
  if branch_filter is not None:
    query = query.filter(orm.ChangeInfo.branch.like(branch_filter))

  query = query.order_by(orm.ChangeInfo.poll_id.desc(),
                         orm.ChangeInfo.priority.asc(),
                         orm.ChangeInfo.queue_time.asc())
  count = query.count()

  if offset is not None and offset > 0:
    query = query.offset(offset)

  if limit is not None and limit > 0:
    query = query.limit(limit)

  return count, [common.ChangeInfo(**ci_sql.as_dict()) for ci_sql in query]


def get_history(sql, project_filter, branch_filter, offset, limit):
  """
  Return json serializable list of MergeStatus dictionaries for available
  merge requests matching the given project and branch filters (as regular
  expressions)

  TODO(josh): filter out any which are IN_PROGRESS?
  """
  query = sql.query(orm.MergeStatus)

  if project_filter is not None:
    query = query.filter(orm.MergeStatus.project.like(project_filter))
  if branch_filter is not None:
    query = query.filter(orm.MergeStatus.branch.like(branch_filter))

  query = query.order_by(orm.MergeStatus.end_time.desc())
  count = query.count()

  if offset > 0:
    query = query.offset(offset)

  if limit > 0:
    query = query.limit(limit)

  return dict(count=count, result=[ms_sql.as_dict() for ms_sql in query])


def sync_account_db(gerrit, sql):
  """
  Synchronize local account table to gerrit account table
  """

  page_size = 25
  for page_idx in range(10000):
    offset = page_idx * page_size
    json_list = gerrit.get('accounts/?start={offset}&n={page_size}&o=DETAILS'
                           .format(offset=offset, page_size=page_size))
    for ai_json in json_list:
      query = (sql
               .query(orm.AccountInfo)
               .filter(orm.AccountInfo.rid == ai_json['_account_id']))
      if query.count() > 0:
        for ai_sql in query:
          ai_sql.name = ai_json.get('name', '<none>')
          ai_sql.email = ai_json.get('email', '<none>')
          ai_sql.username = ai_json.get('username', '<none>')
          sql.commit()
          break
      else:
        kwargs = dict(ai_json)
        kwargs['rid'] = kwargs.pop('_account_id')
        for key in ['name', 'email', 'username']:
          if key not in kwargs:
            kwargs[key] = '<none>'

        ai_sql = orm.AccountInfo(**kwargs)
        sql.add(ai_sql)
        sql.commit()

    if len(json_list) < 1 or '_more_accounts' not in json_list[-1]:
      break


def migrate_db(gerrit, input_path, from_version, output_path, to_version):
  """
  Migrate a database from one schema version to another
  """

  if from_version == '0.1.0' and to_version == '0.2.0':
    migrate_db_v0p1p0_to_v0p2p0(gerrit, input_path, output_path)
  elif from_version == '0.2.0' and to_version == '0.2.1':
    migrate_db_v0p2p0_to_v0p2p1(input_path, output_path)


def migrate_db_v0p2p0_to_v0p2p1(input_path, output_path):
  """
  Split merge_history into merge_history and merge_changes.
  """

  tmp_path = input_path + '.mq_migration'

  if os.path.exists(tmp_path):
    logging.info('Removing stale temporary %s', tmp_path)
    os.remove(tmp_path)

  logging.info('Copying %s to %s', input_path, tmp_path)
  shutil.copyfile(input_path, tmp_path)

  logging.info('Creating merge_changes table')
  import sqlite3
  conn = sqlite3.connect(tmp_path)
  cur = conn.cursor()
  cur.execute('SELECT name FROM SQLITE_MASTER WHERE type="index"'
              ' AND tbl_name="merge_history"')
  indices = [row[0] for row in cur]
  for index in indices:
    logging.info('Dropping index %s from merge_history', index)
    cur.execute('DROP INDEX {}'.format(index))
  cur.execute('ALTER TABLE merge_history RENAME TO merge_history_v0p2p0')
  conn.commit()
  conn.close()

  try:
    os.remove(output_path)
  except OSError:
    pass

  os.rename(tmp_path, output_path)

  logging.info('Migrating rows')
  sql = orm.init_sql('sqlite:///{}'.format(output_path))()

  prev_migration_query = (sql.query(orm.MergeStatus)
                          .order_by(orm.MergeStatus.rid.desc())
                          .limit(1))
  source_query = sql.query(orm.MergeStatusV0p2p0)

  if prev_migration_query.count() > 0:
    prev_migration_last = -1
    for prev_status in prev_migration_query:
      prev_migration_last = prev_status.rid
      break

    logging.info('Detected previous migration, will migrate increment'
                 ' starting at row id %d', prev_migration_last)
    source_query = source_query.filter(orm.MergeStatusV0p1p0.id >
                                       prev_migration_last)

  last_print_time = 0
  query_count = source_query.count()
  for idx, old_status in enumerate(source_query):
    kwargs = {key: getattr(old_status, key) for key in
              ['rid', 'project', 'branch', 'start_time', 'end_time', 'status']}
    sql.add(orm.MergeStatus(**kwargs))

    kwargs = {key: getattr(old_status, key) for key in
              ['owner_id', 'change_id', 'request_time', 'msg_meta']}
    kwargs['merge_id'] = old_status.rid
    sql.add(orm.MergeChange(**kwargs))

    sql.commit()

    if time.time() - last_print_time > 0.5:
      last_print_time = time.time()
      progress = 100.0 * (idx + 1) / query_count
      sys.stdout.write('{:6d}/{:6d} [{:6.2f}%]\r'
                       .format(idx, query_count, progress))
      sys.stdout.flush()

  sys.stdout.write('{:6d}/{:6d} [{:6.2f}%]\n'
                   .format(query_count, query_count, 100.0))

  sql.close()

  conn = sqlite3.connect(output_path)
  cur = conn.cursor()
  cur.execute('DROP TABLE merge_history_v0p2p0')
  conn.commit()
  conn.close()


def migrate_db_v0p1p0_to_v0p2p0(gerrit, input_path, output_path):
  """
  Migrate a database from one schema version to another
  """
  tmp_path = input_path + '.mq_migration'

  if os.path.exists(tmp_path):
    logging.info('Removing stale temporary %s', tmp_path)
    os.remove(tmp_path)

  logging.info('Copying %s to %s', input_path, tmp_path)
  shutil.copyfile(input_path, tmp_path)

  logging.info('Renaming old table')
  import sqlite3
  conn = sqlite3.connect(tmp_path)
  cur = conn.cursor()
  cur.execute('ALTER TABLE merge_history RENAME TO merge_history_v0p1p0')
  conn.commit()
  conn.close()

  logging.info('Migrating rows')
  source_sql = orm.init_sql('sqlite:///{}'.format(tmp_path))()
  dest_sql = orm.init_sql('sqlite:///{}'.format(output_path))()

  prev_migration_query = (dest_sql.query(orm.MergeStatus)
                          .order_by(orm.MergeStatus.rid.desc())
                          .limit(1))

  source_query = source_sql.query(orm.MergeStatusV0p1p0)

  if prev_migration_query.count() > 0:
    prev_migration_last = -1
    for prev_status in prev_migration_query:
      prev_migration_last = prev_status.rid
      break

    logging.info('Detected previous migration, will migrate increment'
                 ' starting at row id %d', prev_migration_last)
    source_query = source_query.filter(orm.MergeStatusV0p1p0.id >
                                       prev_migration_last)

  last_print_time = 0
  query_count = source_query.count()
  for idx, old_status in enumerate(source_query):
    if old_status.result == orm.StatusKey.IN_PROGRESS.value:
      status = orm.StatusKey.CANCELED.value
    else:
      status = old_status.result

    changeinfo = None
    msg_meta = {}
    max_tries = 10
    sleep_duration = 2

    for try_idx in range(max_tries):
      try:
        if not changeinfo:
          changeinfo = gerrit.get_change(old_status.change_id)
        if not msg_meta:
          msg_meta = gerrit.get_message_meta(old_status.change_id,
                                             changeinfo.current_revision)
        break
      except requests.RequestException:
        logging.warn('Failed to poll gerrit for change %s %d/%d',
                     old_status.change_id, try_idx, max_tries)
        time.sleep(sleep_duration)

    if changeinfo is not None:
      owner_id = changeinfo.owner.account_id
    else:
      owner_id = -1
    new_status = orm.MergeStatusV0p2p0(rid=old_status.id,
                                       project='aircam',
                                       branch=old_status.target_branch,
                                       owner_id=owner_id,
                                       change_id=old_status.change_id,
                                       request_time=old_status.request_time,
                                       start_time=old_status.start_time,
                                       end_time=old_status.end_time,
                                       msg_meta=json.dumps(msg_meta),
                                       status=status)
    dest_sql.add(new_status)
    dest_sql.commit()

    if time.time() - last_print_time > 0.5:
      last_print_time = time.time()
      progress = 100.0 * (idx + 1) / query_count
      sys.stdout.write('{:6d}/{:6d} [{:6.2f}%]\r'
                       .format(idx, query_count, progress))
      sys.stdout.flush()

  sys.stdout.write('{:6d}/{:6d} [{:6.2f}%]\n'
                   .format(query_count, query_count, 100.0))
  os.remove(tmp_path)


MISSING_IDS_QUERY = """
SELECT DISTINCT(owner_id)
  FROM merge_history LEFT JOIN account_info
    ON owner_id = account_info.rid
  WHERE account_info.rid is NULL
  ORDER BY owner_id ASC
"""


def fetch_missing_account_info(gerrit, db_path):
  """
  Fetch any account info from gerrit that is missing in the database
  """

  import sqlite3

  logging.info('Querying missing ids')
  sql = orm.init_sql('sqlite:///{}'.format(db_path))()
  conn = sqlite3.connect(db_path)
  cur = conn.cursor()
  cur.execute(MISSING_IDS_QUERY)

  missing_ids = [row[0] for row in cur]
  num_missing = len(missing_ids)
  conn.close()

  logging.info('Fetching %d account infos from gerrit', num_missing)
  last_print_time = 0
  for idx, owner_id in enumerate(missing_ids):
    if time.time() - last_print_time > 1:
      last_print_time = time.time()
      progress = 100.0 * (idx + 1) / num_missing
      sys.stdout.write('{:6d}/{:6d} [{:6.2f}%]\r'
                       .format(idx, num_missing, progress))
      sys.stdout.flush()

    try:
      ai_json = gerrit.get('accounts/{}'.format(owner_id))
      ai_obj = common.AccountInfo(**ai_json)
      add_or_update_account_info(sql, ai_obj)
    except requests.RequestException:
      logging.warn('Failed to get account info for owner_id=%d', owner_id)
      continue
    except ValueError:
      logging.warn('Malformed json for owner_id=%d', owner_id)
      continue

  sys.stdout.write('{:6d}/{:6d} [{:6.2f}%]\n'
                   .format(num_missing, num_missing, 100.0))
  sql.close()


def gzip_old_logs(srcdir, destdir):
  """
  For any logfiles that are not already gzipped, gzip them and then create a
  zero sized stub.
  """

  last_print_time = 0
  max_num_logs = 1000000
  gzip_jobs = 0

  for idx in range(max_num_logs):
    if time.time() - last_print_time > 1.0:
      last_print_time = time.time()
      progress = 100.0 * (idx + 1) / max_num_logs
      sys.stdout.write('{:6d}/{:6d} [{:6.2f}%] ({:6d})\r'
                       .format(idx, max_num_logs, progress, gzip_jobs))
      sys.stdout.flush()

    for extension in ['log', 'stderr', 'stdout']:
      filename = '{:06d}.{}'.format(idx, extension)
      logpath = os.path.join(srcdir, filename)
      stub_path = os.path.join(destdir, filename)
      gzip_path = os.path.join(destdir, filename + '.gz')
      if os.path.exists(logpath) and not os.path.exists(gzip_path):
        gzip_jobs += 1
        with open(gzip_path, 'w') as outfile:
          subprocess.check_call(['gzip', '--stdout', logpath], stdout=outfile)
        with open(stub_path, 'w') as _:
          pass

  sys.stdout.write('{:6d}/{:6d} [{:6.2f}%] ({:6d})\n'
                   .format(max_num_logs, max_num_logs, 100.0, gzip_jobs))


def path_prefix_in(query_path, prefix_list):
  query_parts = query_path.split('/')

  for prefix in prefix_list:
    prefix_parts = prefix.split('/')
    if prefix_parts[:len(query_parts)] == query_parts:
      return True
  return False


# this struct will be passed as a ponter,
# so we don't have to worry about the right layout
class dl_phdr_info(ctypes.Structure):  # pylint: disable=invalid-name
  _fields_ = [
      ('padding0', ctypes.c_void_p),  # ignore it
      ('dlpi_name', ctypes.c_char_p),  # ignore the reset
  ]


class LibListStore(object):

  def __init__(self):
    self.liblist = []

  def handle_libinfo(self, info, size, data):  # pylint: disable=unused-argument
    self.liblist.append(info.contents.dlpi_name)
    return 0


def get_loaded_libraries():
  """
  Return a list of file paths for libraries which are loaded into the current
  interpreter process
  """

  # NOTE(josh): c_void_p changed to c_char_p
  callback_t = ctypes.CFUNCTYPE(ctypes.c_int,
                                ctypes.POINTER(dl_phdr_info),
                                ctypes.POINTER(ctypes.c_size_t),
                                ctypes.c_char_p)

  dl_iterate_phdr = ctypes.CDLL('libc.so.6').dl_iterate_phdr

  # NOTE(josh): c_void_p replaced with c_char_p
  dl_iterate_phdr.argtypes = [callback_t, ctypes.c_char_p]
  dl_iterate_phdr.restype = ctypes.c_int

  list_store = LibListStore()
  dl_iterate_phdr(callback_t(list_store.handle_libinfo), "dummy")
  return list_store.liblist


def get_watch_manifest(ignore_prefixes=None):
  """
  Return a list of file paths and mtimes for files that should trigger a restart
  if they change.
  """

  if ignore_prefixes is None:
    ignore_prefixes = []

  zipfiles_on_path = []
  for component in sys.path:
    if component and zipfile.is_zipfile(component):
      realpath_to_zip = os.path.realpath(component)
      zipfiles_on_path.append((realpath_to_zip,
                               os.path.getmtime(realpath_to_zip)))

  zipfiles = [zfile for (zfile, _) in zipfiles_on_path]
  module_files = []
  for _, module in sys.modules.items():
    if hasattr(module, '__file__'):
      realpath_to_module = os.path.realpath(module.__file__)

      if path_prefix_in(realpath_to_module, zipfiles):
        logging.info('Skipping zipfile module %s', realpath_to_module)
        continue
      elif path_prefix_in(realpath_to_module, ignore_prefixes):
        logging.info('Skipping ignored module %s', realpath_to_module)
        continue
      else:
        module_files.append((realpath_to_module,
                             os.path.getmtime(realpath_to_module)))

  so_files = []
  for so_path in get_loaded_libraries():
    realpath_to_so = os.path.realpath(so_path)
    if path_prefix_in(realpath_to_module, ignore_prefixes):
      logging.info('Skipping ignored so %s', realpath_to_so)
      continue
    else:
      so_files.append((realpath_to_so, os.path.getmtime(realpath_to_so)))

  return sorted(zipfiles_on_path + module_files + so_files)


def get_changelist(manifest):
  """
  Given a list of (fullpath, mtime) return a list of paths whose current mtime
  is newer than the manifest mtime.
  """

  changelist = []
  for fullpath, mtime in manifest:
    if os.path.getmtime(fullpath) - mtime > 0.1:
      changelist.append(fullpath)

  return changelist


def get_real_argv():
  """
  Return the actual command line
  """

  with open('/proc/self/cmdline', 'r') as infile:
    return infile.read().split('\0')[:-1]


def restart_if_modified(watch_manifest, pidfile_path):
  """
  Restart the process if any file in the manifest has changed
  """

  changelist = get_changelist(watch_manifest)
  if changelist:
    logging.info('Detected a sourcefile change: \n  '
                 + '\n  '.join(changelist))
    argv = get_real_argv()
    os.remove(pidfile_path)
    os.execvp(sys.executable, argv)

class ZipFileLoader(jinja2.BaseLoader):
  """
  Implements a template loader which reads templates from a zipfile
  """

  def __init__(self, zipfile_path, base_directory):
    self.zipf = zipfile.ZipFile(zipfile_path)
    self.basedir = base_directory

  def __del__(self):
    self.zipf.close()

  def get_source(self, environment, template):
    try:
      fullpath = '{}/{}'.format(self.basedir, template)
      with self.zipf.open(fullpath) as fileobj:
        source = fileobj.read()
    except IOError:
      raise jinja2.TemplateNotFound(template,
                                    message='Fullpath: {}'.format(fullpath))

    return (source, None, lambda: False)


def render_templates(config, outdir):
  """
  Render jinja2 templates into the specified documentroot
  """

  pardir = os.path.dirname(__file__)
  pardir = os.path.dirname(pardir)
  logging.info('pardir: %s', pardir)

  if zipfile.is_zipfile(pardir):
    logging.info('reading data from zipfile')
    loader = ZipFileLoader(pardir, 'gerrit_mq/templates')
  else:
    logging.info('reading data from package directory')
    loader = jinja2.PackageLoader('gerrit_mq', 'templates')

  env = jinja2.Environment(loader=loader)

  for page in ['daemon', 'detail', 'history', 'index', 'queue']:
    template_name = '{}.html.tpl'.format(page)
    template = env.get_template(template_name)
    outpath = os.path.join(outdir, '{}.html'.format(page))
    with open(outpath, 'w') as outfile:
      outfile.write(template.render())  # pylint: disable=no-member
      outfile.write('\n')

  script_path = 'gerrit_mq/templates/script.js.tpl'
  style_path = 'gerrit_mq/templates/style.css'

  if zipfile.is_zipfile(pardir):
    with zipfile.ZipFile(pardir) as zfile:
      with zfile.open(script_path) as infile:
        js_lines = infile.readlines()
      with zfile.open(style_path) as infile:
        style_content = infile.read()
  else:
    with open(os.path.join(pardir, script_path)) as infile:
      js_lines = infile.readlines()
    with open(os.path.join(pardir, style_path)) as infile:
      style_content = infile.read()

  outpath = os.path.join(outdir, 'script.js')
  with open(outpath, 'w') as outfile:
    for line in js_lines:
      if line.startswith('var kGerritURL'):
        outfile.write('var kGerritURL = "{}";\n'
                      .format(config['gerrit.rest.url']))
      else:
        outfile.write(line)

  outpath = os.path.join(outdir, 'style.css')
  with open(outpath, 'w') as outfile:
    outfile.write(style_content)



"""Web frontend for inspecting the merge queue."""

import datetime
import cStringIO
import logging
import os

import flask

from gerrit_mq import orm
from gerrit_mq import functions

HTML_ESCAPE_TABLE = {
    "&": "&amp;",
    '"': "&quot;",
    "'": "&apos;",
    ">": "&gt;",
    "<": "&lt;",
}


def html_escape_file(infile, max_bytes):
  """
  Escape special characters so that the content is html-embeddable
  """
  buf = cStringIO.StringIO()
  bytes_read = 0
  for char in iter(lambda: infile.read(1).decode('utf-8', 'replace'), ''):
    # TODO(josh): fix this in a better way
    if ord(char) < 128:
      buf.write(HTML_ESCAPE_TABLE.get(char, char))
    else:
      # TODO(josh): is this OK for utf-8?
      buf.write(char)
    bytes_read += 1
    if bytes_read >= max_bytes:
      break

  # TODO(josh): Can we force this to be unicode? Create test case.
  return bytes_read, buf.getvalue()


def html_escape_tail(filename, max_bytes):
  """
  Escape special characters so that the content is html-embeddable
  """
  with open(filename, 'r') as infile:
    infile.seek(0, 2)  # move the file stream to the end of the file
    size_in_bytes = infile.tell()
    if max_bytes > size_in_bytes:
      infile.seek(0, 0)  # move back to the beginning
    else:
      infile.seek(-max_bytes, 2)
    _, content = html_escape_file(infile, max_bytes)
    return content


def extract_common_args(req_args):
  """
  Pull common query parameter arguments out of a query string
  """

  # TODO(josh): figure out FlaskSqlAlchemy and how to do this right.
  project_filter = req_args.get('project', None)
  branch_filter = req_args.get('branch', None)

  try:
    offset = max(int(req_args['offset']), 0)
  except (KeyError, ValueError):
    offset = 0

  try:
    limit = min(int(req_args['limit']), 500)
  except (KeyError, ValueError):
    limit = 25

  return project_filter, branch_filter, offset, limit


class Webfront(flask.Flask):

  def __init__(self, mq_config, gerrit, sql_factory):
    super(Webfront, self).__init__('gerrit_mq')
    self.debug = mq_config['webfront.flask_debug']
    self.gerrit = gerrit
    self.sql_factory = sql_factory
    self.mq_config = mq_config
    self.secret_key = mq_config['webfront.secret_key']

    self.add_url_rule('/gmq/cancel_merge', 'cancel_merge', self.cancel_merge)
    self.add_url_rule('/gmq/get_active_merge_status', 'get_active_merge_status',
                      self.get_active_merge_status)
    self.add_url_rule('/gmq/get_queue', 'get_queue', self.get_queue)
    self.add_url_rule('/gmq/get_history', 'get_history', self.get_history)
    self.add_url_rule('/gmq/get_merge_status', 'get_merge_status',
                      self.get_merge_status)
    self.add_url_rule('/gmq/get_daemon_status', 'get_daemon_status',
                      self.get_daemon_status)
    self.add_url_rule('/gmq/set_daemon_pause', 'set_daemon_pause',
                      self.set_daemon_pause)

    logging.info('Initialized webfront app')

  def get_queue(self):
    """
    Return json-encoded list of ChangeInfo objects for the current queue.

    Query params:
    `project` : SQL `LIKE` expression for projects to match
    `branch` : SQL `LIKE` expression for branches to match
    `offset` : start offset for pagination
    `limit` : maximum number of records to return
    """

    project_filter, branch_filter, offset, limit \
        = extract_common_args(flask.request.args)

    sql = self.sql_factory()
    count, result_list = functions.get_queue(sql, project_filter, branch_filter,
                                             offset, limit)
    sql.close()
    return flask.jsonify(count=count,
                         result=[ci.as_dict() for ci in result_list])

    # request_queue = [ci.as_dict() for ci in self.gerrit.get_merge_requests()]
    # return flask.jsonify(count=len(request_queue), result=request_queue)

  def get_history(self):
    """
    Return json-encoded list of MergeStatus objects.

    Query params:
      `project` : SQL `LIKE` expression for projects to match
      `branch` : SQL `LIKE` expression for branches to match
      `offset` : start offset for pagination
      `limit` : maximum number of records to return
    """
    project_filter, branch_filter, offset, limit \
        = extract_common_args(flask.request.args)

    sql = self.sql_factory()
    query = sql.query(orm.MergeStatus)

    if project_filter is not None:
      query = query.filter(orm.MergeStatus.project.like(project_filter))
    if branch_filter is not None:
      query = query.filter(orm.MergeStatus.branch.like(branch_filter))

    query = query.order_by(orm.MergeStatus.rid.desc())
    count = query.count()

    if offset > 0:
      query = query.offset(offset)

    if limit > 0:
      query = query.limit(limit)

    result = []
    for merge_sql in list(query):
      merge_json = merge_sql.as_dict()
      query = (sql.query(orm.MergeChange)
               .join(orm.MergeChange.owner)
               .filter(orm.MergeChange.merge_id == merge_sql.rid))
      merge_json['changes'] = []
      for change_sql in query:
        merge_json['changes'].append(change_sql.as_dict())
      result.append(merge_json)

    response = flask.jsonify(dict(count=count,
                                  result=result))
    sql.close()
    return response

  def get_merge_status(self):
    """
    Return json-encoded MergeStatus for a single merge

    Query params:
      `rid` : row id of the status to retrieve
    """
    sql = self.sql_factory()
    if 'rid' in flask.request.args:
      try:
        query_rid = int(flask.request.args['rid'])
      except (KeyError, ValueError):
        response = flask.jsonify({'status': 'ERROR',
                                  'reason': 'invalid rid'})
        response.status_code = 400
        sql.close()
        return response

      query = (sql
               .query(orm.MergeStatus)
               .filter(orm.MergeStatus.rid == query_rid))
    else:
      query = (sql
               .query(orm.MergeStatus).order_by(orm.MergeStatus.rid.desc())
               .limit(1))

    if query.count() < 1:
      sql.close()
      response = flask.jsonify({'status': 'ERROR',
                                'reason': "rid doesn't exist in db"})
      response.status_code = 404
      return response

    record_sql = query.first()
    record_json = record_sql.as_dict()
    record_json['changes'] = []

    query = (sql.query(orm.MergeChange)
             .filter(orm.MergeChange.merge_id == query_rid)
             .order_by(orm.MergeChange.request_time))
    for change_sql in query:
      record_json['changes'].append(change_sql.as_dict())

    sql.close()

    return flask.jsonify(record_json)

  def get_active_merge_status(self):
    """
    Return json-encoded MergeStatus for a single merge

    Query params:
      `rid` : row id of the status to retrieve
    """
    sql = self.sql_factory()
    query = (sql
             .query(orm.MergeStatus)
             # .filter(orm.MergeStatus.status
             #         == orm.StatusKey.IN_PROGRESS.value)
             .order_by(orm.MergeStatus.rid.desc())
             .limit(1))

    if query.count() < 1:
      sql.close()
      return flask.jsonify({})

    record_sql = query.first()
    record_json = record_sql.as_dict()
    record_json['changes'] = []

    query = (sql.query(orm.MergeChange)
             .filter(orm.MergeChange.merge_id == record_sql.rid)
             .order_by(orm.MergeChange.request_time))
    for change_sql in query:
      record_json['changes'].append(change_sql.as_dict())

    sql.close()
    return flask.jsonify(record_json)

  def cancel_merge(self):
    """
    Query params:
      `rid` : row id of the status to retrieve
    """
    sql = self.sql_factory()
    try:
      query_rid = int(flask.request.args['rid'])
    except (KeyError, ValueError):
      response = flask.jsonify({'status': 'ERROR',
                                'reason': 'invalid rid'})
      response.status_code = 400
      sql.close()
      return response

    query = (sql
             .query(orm.Cancellation)
             .filter(orm.Cancellation.rid == query_rid))
    if query.count() > 0:
      sql.close()
      return flask.jsonify({'status': 'SUCCESS',
                            'note': 'Already Canceled in DB'})

    row = orm.Cancellation(rid=query_rid, when=datetime.datetime.utcnow(),
                           who='Webfront')
    sql.add(row)
    sql.commit()
    sql.close()
    return flask.jsonify({'status': 'SUCCESS'})

  def get_daemon_status(self):
    """
    Return json-encoded info about the running daemon.
    """
    pausefile_path = self.mq_config['daemon.offline_sentinel_path']
    pidfile_path = self.mq_config['daemon.pidfile_path']
    daemon_pid = -1
    if os.path.exists(pidfile_path):
      try:
        with open(pidfile_path, 'r') as infile:
          daemon_pid = int(infile.read().strip())
      except (ValueError, IOError, OSError):
        pass

    status = {
        'alive': os.path.exists('/proc/{}/stat'.format(daemon_pid)),
        'paused': os.path.exists(pausefile_path),
        'pid': daemon_pid,
    }

    return flask.jsonify(status)

  def set_daemon_pause(self):
    """
    Touch or remove the pause sentinel
    """
    pausefile_path = self.mq_config['daemon.offline_sentinel_path']

    try:

      query_value = (flask.request.args.get('value', 'true') == 'true')
    except (KeyError, ValueError):
      query_value = True

    if query_value:
      with open(pausefile_path, 'w') as _:
        pass
    else:
      os.remove(pausefile_path)

    return self.get_daemon_status()



"""Entry point / launcher for gerrit-mq components."""

from __future__ import print_function
import argparse
import inspect
import io
import json
import logging
import os
import re
import sys
import traceback

import gerrit_mq
from gerrit_mq import common
from gerrit_mq import daemon
from gerrit_mq import functions
from gerrit_mq import orm

def class_to_cmd(name):
  intermediate = re.sub('(.)([A-Z][a-z]+)', r'\1-\2', name)
  return re.sub('([a-z0-9])([A-Z])', r'\1-\2', intermediate).lower()


class Command(object):
  """
  Base class making it a little easier to set up a complex argparse tree by
  specifying features of a command as memebers of a class.
  """

  default_log_level = 'info'

  @staticmethod
  def setup_parser(subparser):
    """
    Configure subparser for this command. Override in subclasses.
    """
    pass

  @classmethod
  def get_cmd(cls):
    """
    Return a string command name formulated by de-camael-casing the class
    name.
    """
    return class_to_cmd(cls.__name__)

  @classmethod
  def add_parser(cls, subparsers):
    """
    Add a subparser to the list of subparsers, and then call the classmethod
    to configure that subparser.
    """

    subparser = subparsers.add_parser(cls.get_cmd(), help=cls.__doc__)
    cls.setup_parser(subparser)

  @classmethod
  def run_args(cls, config, args):  # pylint: disable=unused-argument
    """
    Override this method to execute the command with the given parsed args.
    """
    raise RuntimeError('run_args unimplemented for object of type {}'
                       .format(getattr(cls, '__name__', '??')))


class PollGerrit(Command):
  """
  Hit gerrit REST and read off the current queue of merge requests. Write that
  to a json file.
  """
  @staticmethod
  def setup_parser(subparser):
    subparser.add_argument('--poll-id', type=int, default=0,
                           help="unique identifier for this poll. Used to "
                                "clear change queue of old changes")

  @classmethod
  def run_args(cls, config, args):
    gerrit = common.GerritRest(**config['gerrit.rest'])
    sql = orm.init_sql(config['db_url'])()

    if args.poll_id == 0:
      args.poll_id = functions.get_next_poll_id(sql)
    functions.poll_gerrit(gerrit, sql, args.poll_id)


class GetQueue(Command):
  """
  Retrieve the currently cached queue in json format
  """
  @staticmethod
  def setup_parser(subparser):
    subparser.add_argument('-p', '--project-filter', default=None,
                           help="which project to select queue items for")
    subparser.add_argument('-b', '--branch-filter', default=None,
                           help="regex to filter branches")
    subparser.add_argument('--offset', type=int, default=0,
                           help="offset for SQL query")
    subparser.add_argument('--limit', type=int, default=0,
                           help="maximum rows to return from SQL query")

  @classmethod
  def run_args(cls, config, args):
    session_factory = orm.init_sql(config['db_url'])
    _, queue = functions.get_queue(session_factory(), args.project_filter,
                                   args.branch_filter, args.offset, args.limit)
    queue = [item.as_dict() for item in queue]
    json.dump(queue, sys.stdout, indent=2, separators=(',', ': '))
    sys.stdout.write('\n')


class GetNext(Command):
  """
  Retrieve the next merge request.
  """
  @staticmethod
  def setup_parser(subparser):
    subparser.add_argument('-p', '--project-filter', default=None,
                           help="which project to select queue items for")
    subparser.add_argument('-b', '--branch-filter', default=None,
                           help="regex to filter branches")

  @classmethod
  def run_args(cls, config, args):
    session_factory = orm.init_sql(config['db_url'])
    queue = functions.get_queue(session_factory(), args.project_filter,
                                args.branch_filter, 0, -1)
    json.dump(queue, sys.stdout, indent=2, separators=(',', ': '))
    sys.stdout.write('\n')


class Webfront(Command):
  """
  Start the merge-queue master service.
  """

  default_log_level = 'warn'

  @classmethod
  def run_args(cls, config, args):
    gerrit = common.GerritRest(**config['gerrit.rest'])
    session_factory = orm.init_sql(config['db_url'])

    from gerrit_mq import webfront
    app = webfront.Webfront(config, gerrit, session_factory)
    app.run(**config['webfront.listen'])


class Daemon(Command):
  """
  Execute the daemon process.
  """

  @classmethod
  def run_args(cls, config, args):
    # we need to attempt to create directories before starting the logger
    # because one of the directories s the log directory
    # TODO(josh): implement this
    # common.create_directories(config)

    # We'll add a handler which puts log events in an actual file for review as
    # needed. We'll put the log file on a rotation where each log may grow up
    # to 1 megabyte with up to 10 backups
    filelog = logging.handlers.RotatingFileHandler(
        '{}/app.log'.format(config['log_path']),
        maxBytes=int(1e6), backupCount=10)

    # We'll add a timestamp to the format for this log
    format_str = ('%(asctime)s %(levelname)-8s %(filename)s [%(lineno)-3s] '
                  ': %(message)s')
    filelog.setFormatter(logging.Formatter(format_str))
    logging.getLogger('').addHandler(filelog)

    gerrit = common.GerritRest(**config['gerrit.rest'])
    session_factory = orm.init_sql(config['db_url'])
    app = daemon.MergeDaemon(config, gerrit, session_factory())

    watch_manifest = functions.get_watch_manifest()
    realpath_config = os.path.realpath(args.config_path)
    config_manifest = [(realpath_config, os.path.getmtime(realpath_config))]

    watch_manifest = sorted(watch_manifest + config_manifest)

    exit_code = 1
    try:
      app.run(watch_manifest)
      exit_code = 0
    except:  # pylint: disable=bare-except
      logging.exception('Exiting daemon due to uncought exception')

    sys.exit(exit_code)




class RenderTemplates(Command):
  """
  Render jinja2 templates into full html files.
  """

  @staticmethod
  def setup_parser(subparser):
    subparser.add_argument('outdir', nargs='?', default=None,
                           help="where to write the rendered files")

  @classmethod
  def run_args(cls, config, args):  # pylint: disable=unused-argument
    if args.outdir is None:
      args.outdir = config['webfront.pagedir_path']

    try:
      os.makedirs(args.outdir)
    except OSError:
      pass

    functions.render_templates(config, args.outdir)

class SyncAccountTable(Command):
  """
  Fetch account table from gerrit and store locally
  """

  @classmethod
  def run_args(cls, config, args):  # pylint: disable=unused-argument
    gerrit = common.GerritRest(**config['gerrit.rest'])
    session_factory = orm.init_sql(config['db_url'])
    functions.sync_account_db(gerrit, session_factory())


class MigrateDatabase(Command):
  """
  Migrate a database from one schema to another
  """

  @staticmethod
  def setup_parser(subparser):
    db_versions = ['0.1.0', '0.2.0', '0.2.1']
    subparser.add_argument('input_path',
                           help='Path to the source database')
    subparser.add_argument('output_path',
                           help='Path to the destination database')
    subparser.add_argument('-f', '--from-version', required=True,
                           choices=db_versions)
    subparser.add_argument('-t', '--to-version', required=True,
                           choices=db_versions)

  @classmethod
  def run_args(cls, config, args):
    gerrit = common.GerritRest(**config['gerrit.rest'])
    functions.migrate_db(gerrit, args.input_path, args.from_version,
                         args.output_path, args.to_version)


class FetchMissingAccountInfo(Command):
  """
  Download AccountInfo from gerrit
  """

  @staticmethod
  def setup_parser(subparser):
    subparser.add_argument('db_path', help='path to the database')

  @classmethod
  def run_args(cls, config, args):
    gerrit = common.GerritRest(**config['gerrit.rest'])
    functions.fetch_missing_account_info(gerrit, args.db_path)


class GzipOldLogs(Command):
  """
  Gzip files in an old log directory
  """
  @staticmethod
  def setup_parser(subparser):
    subparser.add_argument('source_dir',
                           help='Directory containing old unzipped logs')
    subparser.add_argument('dest_dir',
                           help='Directory containing new gzipped logs')

  @classmethod
  def run_args(cls, config, args):
    functions.gzip_old_logs(args.source_dir, args.dest_dir)


def iter_command_classes():
  """
  Return a list of all Command subclasses in this file.
  """

  for _, cmd_class in globals().iteritems():
    if (inspect.isclass(cmd_class)
            # pylint: disable=bad-continuation
            and issubclass(cmd_class, Command)
            and cmd_class is not Command):
      yield cmd_class


def main(argv):
  parser = argparse.ArgumentParser(prog="gerrit-mq", description=__doc__)
  parser.add_argument('-c', '--config-path',
                      default=os.path.expanduser('~/.gerrit-mq.py'),
                      help='path to config file')
  parser.add_argument('-l', '--log-level', default=None,
                      choices=['debug', 'info', 'warning', 'error'])
  parser.add_argument('-v', '--version', action='version',
                      version=gerrit_mq.VERSION)
  subparsers = parser.add_subparsers(dest='command', metavar='CMD')
  commands = [init() for init in iter_command_classes()]

  for command in commands:
    command.add_parser(subparsers)

  try:
    import argcomplete
    argcomplete.autocomplete(parser)
  except ImportError:
    pass

  args = parser.parse_args(argv)

  # set up main logger, which logs everything. We'll leave this one logging
  # to the console
  format_str = '%(levelname)-8s %(filename)s [%(lineno)-3s] : %(message)s'
  if args.log_level is None:
    log_level = 'info'
  else:
    log_level = args.log_level

  logging.basicConfig(level=getattr(logging, log_level.upper()),
                      format=format_str,
                      datefmt='%Y-%m-%d %H:%M:%S',
                      filemode='w')

  assert os.path.exists(args.config_path), (
      "The requested config file {} does not exist".format(args.config_path))

  try:
    globals_ = {}
    with io.open(args.config_path, 'r', encoding='utf-8') as infile:
      exec(infile.read(), globals_) # pylint: disable=W0122

  except:  # pylint: disable=bare-except
    traceback.print_exc()
    sys.stderr.write('Failed to execute config file\n')
    return 1

  config = common.ConfigDict(globals_)
  for command in commands:
    if args.command == command.get_cmd():
      if args.log_level is None:
        logging.getLogger('').setLevel(command.default_log_level.upper())
      return command.run_args(config, args)


if __name__ == '__main__':
  sys.exit(main(sys.argv[1:]))



import io
from setuptools import setup

GITHUB_URL = 'https://github.com/cheshirekow/gerrit_mq'

VERSION = None
with io.open('gerrit_mq/__init__.py', 'r', encoding='utf-8') as infile:
  for line in infile:
    line = line.strip()
    if line.startswith('VERSION ='):
      VERSION = line.split('=', 1)[1].strip().strip("'")

assert VERSION is not None

with io.open('README.rst', encoding='utf-8') as infile:
  long_description = infile.read()

setup(
    name='gerrit_mq',
    packages=['gerrit_mq'],
    version=VERSION,
    description="Gerrit merge serializer",
    long_description=long_description,
    author='Josh Bialkowski',
    author_email='josh.bialkowski@gmail.com',
    url=GITHUB_URL,
    download_url='{}/archive/{}.tar.gz'.format(GITHUB_URL, VERSION),
    keywords=['gerrit', 'continuous-integration'],
    classifiers=[],
    entry_points={
        'console_scripts': [
            'gerrit-mq=gerrit_mq.__main__:main'
        ],
    },
    install_requires=[
        'enum',
        'Flask',
        'httplib2',
        'jinja2',
        'pygerrit2',
        'requests',
        'sqlalchemy',
    ]
)
